#!/usr/bin/env python
"""Test RDFStruct implementations."""

import base64
import random

from absl import app
from absl.testing import absltest

from google.protobuf import any_pb2
from google.protobuf import wrappers_pb2
from grr_response_core.lib import rdfvalue
from grr_response_core.lib import type_info
from grr_response_core.lib.rdfvalues import client as rdf_client
from grr_response_core.lib.rdfvalues import client_fs as rdf_client_fs
from grr_response_core.lib.rdfvalues import client_network as rdf_client_network
from grr_response_core.lib.rdfvalues import client_stats as rdf_client_stats
from grr_response_core.lib.rdfvalues import flows as rdf_flows
from grr_response_core.lib.rdfvalues import paths as rdf_paths
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_core.lib.rdfvalues import test_base as rdf_test_base
from grr_response_proto import tests_pb2
from grr_response_server.rdfvalues import flow_objects as rdf_flow_objects
from grr.test_lib import test_lib

# pylint: mode=test


class TestStructWithManyFields(rdf_structs.RDFProtoStruct):
  """A test struct object."""

  type_description = type_info.TypeDescriptorSet(*[
      rdf_structs.ProtoString(
          name="foobar_%d" % i,
          field_number=i + 1,
          default="string",
          description="A string value",
      )
      for i in range(100)
  ])


class TestStruct(rdf_structs.RDFProtoStruct):
  """A test struct object."""

  type_description = type_info.TypeDescriptorSet(
      rdf_structs.ProtoString(
          name="foobar",
          field_number=1,
          default="string",
          description="A string value",
          labels=[rdf_structs.SemanticDescriptor.Labels.HIDDEN],
      ),
      rdf_structs.ProtoUnsignedInteger(
          name="int", field_number=2, default=5, description="An integer value"
      ),
      rdf_structs.ProtoList(
          rdf_structs.ProtoString(
              name="repeated",
              field_number=3,
              description="A repeated string value",
          )
      ),
      # We can serialize an arbitrary RDFValue. This will be serialized into a
      # binary string and parsed on demand.
      rdf_structs.ProtoRDFValue(
          name="urn",
          field_number=6,
          default=rdfvalue.RDFURN("www.google.com"),
          rdf_type="RDFURN",
          description="An arbitrary RDFValue field.",
      ),
      rdf_structs.ProtoEnum(
          name="type",
          field_number=7,
          enum_name="Type",
          enum={"FIRST": 1, "SECOND": 2, "THIRD": 3},
          default=3,
          description="An enum field",
      ),
      rdf_structs.ProtoFloat(
          name="float",
          field_number=8,
          description="A float number",
          default=1.1,
      ),
  )


# In order to define a recursive structure we must add it manually after the
# class definition.
TestStruct.AddDescriptor(
    rdf_structs.ProtoEmbedded(name="nested", field_number=4, nested=TestStruct),
)

TestStruct.AddDescriptor(
    rdf_structs.ProtoList(
        rdf_structs.ProtoEmbedded(
            name="repeat_nested", field_number=5, nested=TestStruct
        )
    ),
)


class VersionedTestStructV1(rdf_structs.RDFProtoStruct):
  """A test struct object."""

  type_description = type_info.TypeDescriptorSet(
      rdf_structs.ProtoBoolean(
          name="bool1",
          field_number=1,
          default=False,
      )
  )


class VersionedTestStructV2(rdf_structs.RDFProtoStruct):
  """A test struct object."""

  type_description = type_info.TypeDescriptorSet(
      rdf_structs.ProtoBoolean(
          name="bool1",
          field_number=1,
          default=False,
      ),
      rdf_structs.ProtoBoolean(
          name="bool2",
          field_number=2,
          default=False,
      ),
  )


class TestStructWithBool(rdf_structs.RDFProtoStruct):
  """A test struct object."""

  type_description = type_info.TypeDescriptorSet(
      rdf_structs.ProtoBoolean(
          name="foo",
          field_number=1,
          default=False,
      )
  )


class TestStructWithEnum(rdf_structs.RDFProtoStruct):
  """A test struct object."""

  type_description = type_info.TypeDescriptorSet(
      rdf_structs.ProtoEnum(
          name="foo",
          field_number=1,
          enum_name="Foo",
          enum={"ZERO": 0, "ONE": 1, "TWO": 2},
          default=0,
      )
  )


class PartialTest1(rdf_structs.RDFProtoStruct):
  """This is a protobuf with fewer fields than TestStruct."""

  type_description = type_info.TypeDescriptorSet(
      rdf_structs.ProtoUnsignedInteger(name="int", field_number=2),
  )


class DynamicTypeTest(rdf_structs.RDFProtoStruct):
  """A protobuf with dynamic types."""

  type_description = type_info.TypeDescriptorSet(
      rdf_structs.ProtoString(
          name="type",
          field_number=1,
          # By default return the TestStruct proto.
          default="TestStruct",
          description="A string value",
      ),
      rdf_structs.ProtoDynamicEmbedded(
          name="dynamic",
          # The callback here returns the type specified by the type member.
          dynamic_cb=lambda x: rdf_structs.RDFProtoStruct.classes.get(x.type),
          field_number=2,
          description="A dynamic value based on another field.",
      ),
      rdf_structs.ProtoEmbedded(
          name="nested", field_number=3, nested=rdf_client.User
      ),
  )


class DynamicAnyValueTypeTest(rdf_structs.RDFProtoStruct):
  """A protobuf with dynamic types stored in AnyValue messages."""

  type_description = type_info.TypeDescriptorSet(
      rdf_structs.ProtoString(
          name="type", field_number=1, description="A string value"
      ),
      rdf_structs.ProtoDynamicAnyValueEmbedded(
          name="dynamic",
          # The callback here returns the type specified by the type member.
          dynamic_cb=lambda x: rdf_structs.RDFProtoStruct.classes.get(x.type),
          field_number=2,
          description="A dynamic value based on another field.",
      ),
  )


class AnyValueWithoutTypeFunctionTest(rdf_structs.RDFProtoStruct):
  """A protobuf with dynamic types stored in AnyValue messages."""

  type_description = type_info.TypeDescriptorSet(
      rdf_structs.ProtoDynamicAnyValueEmbedded(
          name="dynamic", field_number=1, description="A dynamic value."
      ),
  )


class LateBindingTest(rdf_structs.RDFProtoStruct):
  type_description = type_info.TypeDescriptorSet(
      # A nested protobuf referring to an undefined type.
      rdf_structs.ProtoEmbedded(
          name="nested", field_number=1, nested="UndefinedYet"
      ),
      rdf_structs.ProtoRDFValue(
          name="rdfvalue",
          field_number=6,
          rdf_type="UndefinedRDFValue",
          description="An undefined RDFValue field.",
      ),
      # A repeated late bound field.
      rdf_structs.ProtoList(
          rdf_structs.ProtoRDFValue(
              name="repeated",
              field_number=7,
              rdf_type="UndefinedRDFValue2",
              description="An undefined RDFValue field.",
          )
      ),
  )


class UnionTest(rdf_structs.RDFProtoStruct):
  union_field = "struct_flavor"

  type_description = type_info.TypeDescriptorSet(
      rdf_structs.ProtoEnum(
          name="struct_flavor",
          field_number=1,
          enum_name="Type",
          enum={"FIRST": 1, "SECOND": 2, "THIRD": 3},
          default=3,
          description="An union enum field",
      ),
      rdf_structs.ProtoFloat(
          name="first",
          field_number=2,
          description="A float number",
          default=1.1,
      ),
      rdf_structs.ProtoString(
          name="second",
          field_number=3,
          default="string",
          description="A string value",
      ),
      rdf_structs.ProtoUnsignedInteger(
          name="third",
          field_number=4,
          default=5,
          description="An integer value",
      ),
  )


class RDFStructsTest(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  """Test the RDFStruct implementation."""

  rdfvalue_class = TestStruct

  def CheckRDFValue(self, value, sample):
    # Verify that floats are almost equal, since serialization and conversion
    # of floats is not stable.
    self.assertAlmostEqual(value.float, sample.float)

    # Verify that all other fields are exactly equal.
    value_copy = value.Copy()
    value_copy.float = None

    sample_copy = sample.Copy()
    sample_copy.float = None

    super().CheckRDFValue(value_copy, sample_copy)

  def GenerateSample(self, number=1):
    return self.rdfvalue_class(
        int=number,
        foobar="foo%s" % number,
        urn="www.example.com",
        float=2.3 + number,
    )

  def testDynamicType(self):
    test_pb = DynamicTypeTest()
    # We can not assign arbitrary values to the dynamic field.
    self.assertRaises(ValueError, setattr, test_pb, "dynamic", "hello")

    # Can assign a nested field.
    test_pb.dynamic.foobar = "Hello"
    self.assertIsInstance(test_pb.dynamic, TestStruct)

    # Test serialization/deserialization.
    serialized = test_pb.SerializeToBytes()
    self.assertEqual(DynamicTypeTest.FromSerializedBytes(serialized), test_pb)
    test_struct_class_repr = "'TestStruct'"

    expected_emitted_proto = """\
message DynamicTypeTest {{

  // A string value
  optional string type = 1 [default = {test_struct}];

  // A dynamic value based on another field.
  optional bytes dynamic = 2;
  optional User nested = 3;
}}
""".format(test_struct=test_struct_class_repr)

    # Test proto definition.
    self.assertEqual(DynamicTypeTest.EmitProto(), expected_emitted_proto)

  def testAnyValueWithoutTypeCallback(self):
    test_pb = AnyValueWithoutTypeFunctionTest()

    for value_to_assign in [
        rdfvalue.RDFString("test"),
        rdfvalue.RDFInteger(1234),
        rdfvalue.RDFBytes(b"abc"),
        rdf_flows.GrrStatus(status="WORKER_STUCK", error_message="stuck"),
    ]:
      test_pb.dynamic = value_to_assign
      serialized = test_pb.SerializeToBytes()
      deserialized = AnyValueWithoutTypeFunctionTest.FromSerializedBytes(
          serialized
      )
      self.assertEqual(deserialized, test_pb)
      self.assertEqual(type(deserialized.dynamic), type(value_to_assign))

  def testDynamicAnyValueType(self):
    test_pb = DynamicAnyValueTypeTest(type="TestStruct")
    # We can not assign arbitrary values to the dynamic field.
    self.assertRaises(ValueError, setattr, test_pb, "dynamic", "hello")

    # Can assign a nested field.
    test_pb.dynamic.foobar = "Hello"
    self.assertIsInstance(test_pb.dynamic, TestStruct)

    # Test serialization/deserialization.
    serialized = test_pb.SerializeToBytes()
    self.assertEqual(
        DynamicAnyValueTypeTest.FromSerializedBytes(serialized), test_pb
    )

    # Test proto definition.
    self.assertEqual(
        DynamicAnyValueTypeTest.EmitProto(),
        "message DynamicAnyValueTypeTest {\n\n  "
        "// A string value\n  optional string type = 1;\n\n  "
        "// A dynamic value based on another field.\n  "
        "optional google.protobuf.Any dynamic = 2;\n}\n",
    )

  def testUninitializedDynamicValueIsNonePerDefault(self):
    response = rdf_flow_objects.FlowResponse()  # Do not set payload.
    self.assertIsNone(response.payload)

  def testDynamicAnyValueTypeWithPrimitiveValues(self):
    test_pb = DynamicAnyValueTypeTest(type="RDFString")
    # We can not assign arbitrary values to the dynamic field.
    self.assertRaises(ValueError, setattr, test_pb, "dynamic", 42)

    # Can assign a nested field.
    test_pb.dynamic = rdfvalue.RDFString("Hello")
    self.assertIsInstance(test_pb.dynamic, rdfvalue.RDFString)

    # Test serialization/deserialization.
    serialized = test_pb.SerializeToBytes()
    unserialized = DynamicAnyValueTypeTest.FromSerializedBytes(serialized)
    self.assertEqual(unserialized, test_pb)
    self.assertEqual(unserialized.dynamic, "Hello")

  def testStructDefinition(self):
    """Ensure that errors in struct definitions are raised."""
    # A descriptor without a field number should raise.
    self.assertRaises(
        type_info.TypeValueError, rdf_structs.ProtoEmbedded, name="name"
    )

    # Adding a duplicate field number should raise.
    self.assertRaises(
        type_info.TypeValueError,
        TestStruct.AddDescriptor,
        rdf_structs.ProtoUnsignedInteger(name="int", field_number=2),
    )

    # Adding a descriptor which is not a Proto* descriptor is not allowed for
    # Struct fields:
    self.assertRaises(
        type_info.TypeValueError,
        TestStruct.AddDescriptor,
        type_info.String(name="int"),
    )

  def testRepeatedMember(self):
    tested = TestStruct(int=5)
    tested.foobar = "Hello"
    tested.repeated.Append("Good")
    tested.repeated.Append("Bye")

    for i in range(10):
      tested.repeat_nested.Append(foobar="Nest%s" % i)

    data = tested.SerializeToBytes()

    # Parse it again.
    new_tested = TestStruct.FromSerializedBytes(data)

    # Test the repeated field.
    self.assertLen(new_tested.repeat_nested, 10)
    self.assertEqual(new_tested.repeat_nested[1].foobar, "Nest1")

    # Check that slicing works.
    sliced = new_tested.repeat_nested[3:5]
    self.assertEqual(sliced.__class__, new_tested.repeat_nested.__class__)
    self.assertEqual(
        sliced.type_descriptor, new_tested.repeat_nested.type_descriptor
    )

    self.assertLen(sliced, 2)
    self.assertEqual(sliced[0].foobar, "Nest3")

  def testRepeatedFieldHelperEqualToSequences(self):
    struct = TestStruct(repeated=["a", "b", "c"])

    self.assertNotEqual(struct.repeated, "abc")
    self.assertNotEqual(struct.repeated, {"a", "b", "c"})
    self.assertNotEqual(struct.repeated, {"a": "a", "b": "b", "c": "c"})

    self.assertNotEqual(struct.repeated, ["a"])
    self.assertNotEqual(struct.repeated, ["a", "b"])
    self.assertNotEqual(struct.repeated, ["a", "c", "b"])
    self.assertEqual(struct.repeated, ["a", "b", "c"])

    self.assertNotEqual(struct.repeated, ("a",))
    self.assertNotEqual(struct.repeated, ("a", "b"))
    self.assertNotEqual(struct.repeated, ("a", "c", "b"))
    self.assertEqual(struct.repeated, ("a", "b", "c"))

    self.assertIs(struct.repeated.__eq__(dict()), NotImplemented)
    self.assertIs(struct.repeated.__eq__(set()), NotImplemented)
    self.assertIs(struct.repeated.__eq__(b""), NotImplemented)
    self.assertIs(struct.repeated.__eq__(""), NotImplemented)
    self.assertIs(struct.repeated.__eq__(object()), NotImplemented)

  def testUnknownFields(self):
    """Test that unknown fields are preserved across decode/encode cycle."""
    tested = TestStruct(foobar="hello", int=5)
    tested.nested.foobar = "goodbye"

    self.assertEqual(tested.foobar, "hello")
    self.assertEqual(tested.nested.foobar, "goodbye")

    data = tested.SerializeToBytes()

    # Now unpack using a protobuf with less fields defined.
    reduced_tested = PartialTest1.FromSerializedBytes(data)

    self.assertEqual(reduced_tested.int, 5)
    self.assertRaises(AttributeError, getattr, reduced_tested, "foobar")

    # Re-Serialize using the simple protobuf.
    data2 = reduced_tested.SerializeToBytes()
    decoded_tested = TestStruct.FromSerializedBytes(data2)

    # The foobar field should have been preserved, despite PartialTest1() not
    # understanding it.
    self.assertEqual(decoded_tested.foobar, "hello")

    # Check that nested fields are also preserved.
    self.assertEqual(decoded_tested.nested.foobar, "goodbye")

  def testRDFStruct(self):
    tested = TestStruct()

    # can't set integers for string attributes.
    self.assertRaises(type_info.TypeValueError, setattr, tested, "foobar", 1)

    # This is a string so a string assignment is good:
    tested.foobar = "Hello"
    self.assertEqual(tested.foobar, "Hello")

    # This field must be another TestStruct instance..
    self.assertRaises(ValueError, setattr, tested, "nested", "foo")

    # Its ok to assign a compatible semantic protobuf.
    tested.nested = TestStruct(foobar="nested_foo")

    # Not OK to use the wrong semantic type.
    self.assertRaises(
        ValueError, setattr, tested, "nested", PartialTest1(int=1)
    )

    # Not OK to assign a serialized string - even if it is for the right type -
    # since there is no type checking.
    serialized = TestStruct(foobar="nested_foo").SerializeToBytes()
    self.assertRaises(ValueError, setattr, tested, "nested", serialized)

    # Nested accessors.
    self.assertEqual(tested.nested.foobar, "nested_foo")

    # Test repeated elements:

    # Empty list is ok:
    tested.repeated = []
    self.assertEqual(tested.repeated, [])

    tested.repeated = ["string"]
    self.assertEqual(tested.repeated, ["string"])

    self.assertRaises(
        type_info.TypeValueError, setattr, tested, "repeated", [1, 2, 3]
    )

    # Coercing on assignment. This field is an RDFURN:
    tested.urn = "www.example.com"
    self.assertIsInstance(tested.urn, rdfvalue.RDFURN)

    self.assertEqual(tested.urn, rdfvalue.RDFURN("www.example.com"))

    # Test enums.
    self.assertEqual(tested.type, 3)
    self.assertEqual(tested.type.name, "THIRD")

    tested.type = "FIRST"
    self.assertEqual(tested.type, 1)

    # Check that string assignments are case-insensitive.
    tested.type = "second"
    self.assertEqual(tested.type, 2)
    tested.type = "ThIrD"
    self.assertEqual(tested.type, 3)

    # Non-valid types are rejected.
    self.assertRaises(type_info.TypeValueError, setattr, tested, "type", "Foo")

    # Strings of digits should be accepted.
    tested.type = "2"
    self.assertEqual(tested.type, 2)
    # unicode strings should be treated the same way.
    tested.type = "2"
    self.assertEqual(tested.type, 2)
    # Out of range values are permitted and preserved through serialization.
    tested.type = 4
    self.assertEqual(tested.type, 4)
    serialized_type = str(tested.type).encode("utf-8")
    tested.type = 1
    tested.type = serialized_type
    self.assertEqual(tested.type, 4)

  def testCacheInvalidation(self):
    path = rdf_paths.PathSpec(path="/", pathtype=rdf_paths.PathSpec.PathType.OS)
    for x in "01234":
      path.last.Append(path=x, pathtype=rdf_paths.PathSpec.PathType.OS)

    serialized = path.SerializeToBytes()

    path = rdf_paths.PathSpec.FromSerializedBytes(serialized)

    # At this point the wire format cache is fully populated (since the proto
    # had been parsed). We change a deeply nested member.
    path.last.path = "booo"

    # When we serialize the modified proto we should get the new field
    # serialized. If the cache is not properly invalidated, we will return the
    # old result instead.
    self.assertIn(b"booo", path.SerializeToBytes())

  def testLateBinding(self):
    # The LateBindingTest protobuf is not fully defined.
    self.assertRaises(
        KeyError, LateBindingTest.type_infos.__getitem__, "nested"
    )

    self.assertIn("UndefinedYet", rdfvalue._LATE_BINDING_STORE)

    # We can still use this protobuf
    tested = LateBindingTest()

    # But it does not know about the field yet.
    self.assertRaises(AttributeError, tested.Get, "nested")
    self.assertRaises(AttributeError, LateBindingTest, nested=None)

    # Now define the class. This should resolve the late bound fields and re-add
    # them to their owner protobufs.
    class UndefinedYet(rdf_structs.RDFProtoStruct):
      type_description = type_info.TypeDescriptorSet(
          rdf_structs.ProtoString(
              name="foobar", field_number=1, description="A string value"
          ),
      )

    # The field is now resolved.
    self.assertNotIn("UndefinedYet", rdfvalue._LATE_BINDING_STORE)
    nested_field = LateBindingTest.type_infos["nested"]
    self.assertEqual(nested_field.name, "nested")

    # We can now use the protobuf as normal.
    tested = LateBindingTest()
    tested.nested.foobar = "foobar string"
    self.assertIsInstance(tested.nested, UndefinedYet)

  def testRDFValueLateBinding(self):
    # The LateBindingTest protobuf is not fully defined.
    self.assertRaises(
        KeyError, LateBindingTest.type_infos.__getitem__, "rdfvalue"
    )

    self.assertIn("UndefinedRDFValue", rdfvalue._LATE_BINDING_STORE)

    # We can still use this protobuf
    tested = LateBindingTest()

    # But it does not know about the field yet.
    self.assertRaises(AttributeError, tested.Get, "rdfvalue")
    self.assertRaises(AttributeError, LateBindingTest, rdfvalue="foo")

    # Now define the class. This should resolve the late bound fields and re-add
    # them to their owner protobufs.
    class UndefinedRDFValue(rdfvalue.RDFString):
      pass

    # The field is now resolved.
    self.assertNotIn("UndefinedRDFValue", rdfvalue._LATE_BINDING_STORE)
    rdfvalue_field = LateBindingTest.type_infos["rdfvalue"]
    self.assertEqual(rdfvalue_field.name, "rdfvalue")

    # We can now use the protobuf as normal.
    tested = LateBindingTest(rdfvalue="foo")
    self.assertEqual(type(tested.rdfvalue), UndefinedRDFValue)

  def testRepeatedRDFValueLateBinding(self):
    # The LateBindingTest protobuf is not fully defined.
    self.assertRaises(
        KeyError, LateBindingTest.type_infos.__getitem__, "repeated"
    )

    self.assertIn("UndefinedRDFValue2", rdfvalue._LATE_BINDING_STORE)

    # We can still use this protobuf
    tested = LateBindingTest()

    # But it does not know about the field yet.
    self.assertRaises(AttributeError, tested.Get, "repeated")
    self.assertRaises(AttributeError, LateBindingTest, repeated=["foo"])

    # Now define the class. This should resolve the late bound fields and re-add
    # them to their owner protobufs.
    class UndefinedRDFValue2(rdfvalue.RDFString):
      pass

    # The field is now resolved.
    self.assertNotIn("UndefinedRDFValue2", rdfvalue._LATE_BINDING_STORE)
    rdfvalue_field = LateBindingTest.type_infos["repeated"]
    self.assertEqual(rdfvalue_field.name, "repeated")

    # We can now use the protobuf as normal.
    tested = LateBindingTest(repeated=["foo"])
    self.assertLen(tested.repeated, 1)
    self.assertEqual(tested.repeated[0], "foo")
    self.assertEqual(type(tested.repeated[0]), UndefinedRDFValue2)

  def testRDFValueParsing(self):
    stat = rdf_client_fs.StatEntry.protobuf(st_mode=16877)
    data = stat.SerializeToString()

    result = rdf_client_fs.StatEntry.FromSerializedBytes(data)

    self.assertIsInstance(result.st_mode, rdf_client_fs.StatMode)

  def testDefaults(self):
    """Accessing a field which does not exist returns a default."""
    # An empty protobuf.
    tested = TestStruct()

    # Simple strings.
    self.assertFalse(tested.HasField("foobar"))
    self.assertEqual(tested.foobar, "string")
    self.assertFalse(tested.HasField("foobar"))

    # RDFValues.
    self.assertFalse(tested.HasField("urn"))
    self.assertEqual(tested.urn, "www.google.com")
    self.assertFalse(tested.HasField("urn"))

    # Nested fields: Accessing a nested field will create the nested protobuf.
    self.assertFalse(tested.HasField("nested"))
    self.assertEqual(tested.nested.urn, "www.google.com")
    self.assertTrue(tested.HasField("nested"))
    self.assertFalse(tested.nested.HasField("urn"))

  def testUnionCast(self):
    """Check union structs handling."""
    # An empty protobuf.
    tested_non_union = TestStruct()

    # Raises if union-casting is attempted on non-union proto.
    self.assertRaises(AttributeError, tested_non_union.UnionCast)

    # A proto with a semantic union_field. In this particular proto the chosen
    # union variant is called struct_flavor, but it's arbitrary, we also use
    # eg rule_type referring to the chosen union variant elsewhere. This is
    # determined by the value of union_field.
    tested_union = UnionTest()

    # Returns the default value of the default flavor if both struct_flavor and
    # the flavored field are not set.
    self.assertEqual(tested_union.UnionCast(), 5)

    tested_union.struct_flavor = "FIRST"
    # Returns the default value of the selected flavor if it's not been changed.
    self.assertEqual(tested_union.UnionCast(), 1.1)

    tested_union.struct_flavor = "SECOND"
    # Same as above.
    self.assertEqual(tested_union.UnionCast(), "string")

    tested_union.struct_flavor = "THIRD"
    # Once again.
    self.assertEqual(tested_union.UnionCast(), 5)

    tested_union.third = 1337
    # Returns the set flavor value.
    self.assertEqual(tested_union.UnionCast(), 1337)

    tested_union.first = 1.61803399
    # Raises if there is a flavored field set that doesn't match struct_flavor.
    self.assertRaises(ValueError, tested_union.UnionCast)

  def testClearFieldsWithLabelWorksCorrectly(self):
    t = TestStruct(foobar="foo", int=42)
    t.ClearFieldsWithLabel(rdf_structs.SemanticDescriptor.Labels.HIDDEN)
    self.assertFalse(t.HasField("foobar"))
    self.assertTrue(t.HasField("int"))

  def testClearFieldsWithLabelWorksCorrectlyOnNestedStructures(self):
    t = TestStruct(foobar="foo", int=42)
    t.nested = TestStruct(foobar="bar", int=43)
    t.ClearFieldsWithLabel(rdf_structs.SemanticDescriptor.Labels.HIDDEN)
    self.assertFalse(t.nested.HasField("foobar"))
    self.assertTrue(t.nested.HasField("int"))

  def testConversionToPrimitiveDictNoSerialization(self):
    test_struct = TestStruct(
        foobar="foo",
        int=2,
        repeated=["value0", "value1"],
        nested=TestStruct(int=567),
        repeat_nested=[TestStruct(int=568)],
    )
    expected_dict = {
        "foobar": "foo",
        "int": 2,
        "repeated": ["value0", "value1"],
        "nested": {"int": 567},
        "repeat_nested": [{"int": 568}],
    }
    self.assertEqual(test_struct.ToPrimitiveDict(), expected_dict)

  def testConversionToPrimitiveDictWithSerialization(self):
    test_struct = TestStruct(
        foobar="foo",
        int=2,
        repeated=["value0", "value1"],
        nested=TestStruct(int=567),
        repeat_nested=[TestStruct(int=568)],
    )
    expected_dict = {
        "foobar": "foo",
        "int": "2",  # Serialized
        "repeated": ["value0", "value1"],
        "nested": {"int": "567"},  # Serialized
        "repeat_nested": [{"int": "568"}],  # Serialized
    }
    self.assertEqual(
        test_struct.ToPrimitiveDict(stringify_leaf_fields=True), expected_dict
    )

  def testToPrimitiveDictStringifyBytes(self):
    data = b"\xff\xfe\xff"
    encoded_data = base64.b64encode(data).decode("ascii")

    class FooStruct(rdf_structs.RDFProtoStruct):

      type_description = type_info.TypeDescriptorSet(
          rdf_structs.ProtoBinary(name="data", field_number=1)
      )

    foo_struct = FooStruct(data=data)
    foo_dict = foo_struct.ToPrimitiveDict(stringify_leaf_fields=True)

    self.assertIsInstance(foo_dict["data"], str)
    self.assertEqual(foo_dict["data"], encoded_data)

  def testToPrimitiveDictStringifyRDFBytes(self):
    data = b"\xff\xfe\xff"
    encoded_data = base64.b64encode(data).decode("ascii")

    class BarStruct(rdf_structs.RDFProtoStruct):

      type_description = type_info.TypeDescriptorSet(
          rdf_structs.ProtoRDFValue(
              name="data", field_number=1, rdf_type=rdfvalue.RDFBytes
          )
      )

    bar_struct = BarStruct(data=data)
    bar_dict = bar_struct.ToPrimitiveDict(stringify_leaf_fields=True)

    self.assertIsInstance(bar_dict["data"], str)
    self.assertEqual(bar_dict["data"], encoded_data)

  def _GenerateSampleWithManyFields(self):
    fields = {}
    for _ in range(50):
      key = "foobar_%d" % random.randrange(100)
      fields[key] = key

    sample = TestStructWithManyFields(**fields)

    parsed = TestStructWithManyFields.FromSerializedBytes(
        sample.SerializeToBytes()
    )

    return sample, parsed

  def testSerializationIsStable(self):
    sample1, sample2 = self._GenerateSampleWithManyFields()

    self.assertEqual(sample1.SerializeToBytes(), sample2.SerializeToBytes())

  def testRepeatedFieldHelperComparesToNone(self):
    self.assertNotEqual(TestStruct().repeated, None)

  def testSymmetricEqualityForDisjointFields(self):
    # Prevent a regression, where RDFStruct.__eq__ would be True for instances
    # having the same number of fields, but different fields which are set to
    # the default value.
    self.assertNotEqual(
        rdf_client_stats.CpuSeconds(user_cpu_time=6),
        rdf_client_stats.CpuSeconds(system_cpu_time=0),
    )
    self.assertNotEqual(
        rdf_client_stats.CpuSeconds(system_cpu_time=0),
        rdf_client_stats.CpuSeconds(user_cpu_time=6),
    )

  def testDefaultRepeatedSetterDoesNotChangeEquality(self):
    sample = TestStruct()
    sample.repeated  #  pylint: disable=pointless-statement
    self.assertEqual(sample, TestStruct())

  def testUnsetFieldsHaveSymmetricEqualityWithDefaultValues(self):
    self.assertEqual(TestStruct(repeated=[]), TestStruct())
    self.assertEqual(TestStruct(), TestStruct(repeated=[]))

  def testCanDeserializeOldVersion(self):
    a = VersionedTestStructV2(bool1=True, bool2=True)
    a_ser = a.SerializeToBytes()
    a_old = VersionedTestStructV1.FromSerializedBytes(a_ser)

    with self.assertRaises(AttributeError):
      a_old.Get("bool2")

  def testCanCompareEqualityAcrossVersions(self):
    a = VersionedTestStructV1.FromSerializedBytes(
        VersionedTestStructV2(bool1=True, bool2=True).SerializeToBytes()
    )

    b = VersionedTestStructV1(bool1=True)

    self.assertEqual(a, b)
    self.assertEqual(b, a)


class BooleanToEnumMigrationTest(test_lib.GRRBaseTest):

  def testBooleansAreParsedAsEnums(self):
    a = TestStructWithBool()
    b = TestStructWithEnum.FromSerializedBytes(a.SerializeToBytes())
    self.assertEqual(b.foo, "ZERO")

    a = TestStructWithBool(foo=False)
    b = TestStructWithEnum.FromSerializedBytes(a.SerializeToBytes())
    self.assertEqual(b.foo, "ZERO")

    a = TestStructWithBool(foo=True)
    b = TestStructWithEnum.FromSerializedBytes(a.SerializeToBytes())
    self.assertEqual(b.foo, "ONE")

  def testEnumsAreParsedAsBooleans(self):
    a = TestStructWithEnum()
    b = TestStructWithBool.FromSerializedBytes(a.SerializeToBytes())
    self.assertIs(b.foo, False)

    a = TestStructWithEnum(foo="ZERO")
    b = TestStructWithBool.FromSerializedBytes(a.SerializeToBytes())
    self.assertIs(b.foo, False)

    a = TestStructWithEnum(foo="ONE")
    b = TestStructWithBool.FromSerializedBytes(a.SerializeToBytes())
    self.assertIs(b.foo, True)

  def testNewEnumOptionIsBackwardsCompatibleToTrue(self):
    a = TestStructWithEnum(foo="TWO")
    b = TestStructWithBool.FromSerializedBytes(a.SerializeToBytes())
    self.assertIs(b.foo, True)


class GenericRDFProtoTest(test_lib.GRRBaseTest):

  def testNestedProtobufAssignment(self):
    """Check that we can assign a nested protobuf."""
    container = rdf_client.BufferReference()
    test_path = "C:\\test"
    pathspec = rdf_paths.PathSpec(path=test_path, pathtype=1)

    # Should raise - incompatible RDFType.
    self.assertRaises(
        ValueError, setattr, container, "pathspec", rdfvalue.RDFString("hello")
    )

    # Should raise - incompatible RDFProto type.
    self.assertRaises(
        ValueError,
        setattr,
        container,
        "pathspec",
        rdf_client_fs.StatEntry(st_size=5),
    )

    # Assign directly.
    container.device = pathspec

    self.assertEqual(container.device.path, test_path)

    # Clear the field.
    container.device = None

    # Check the protobuf does not have the field set at all.
    self.assertFalse(container.HasField("pathspec"))

  def testSimpleTypeAssignment(self):
    sample = rdf_client_fs.StatEntry()
    sample.AddDescriptor(
        rdf_structs.ProtoRDFValue(
            name="test",
            field_number=45,
            default=rdfvalue.RDFInteger(0),
            rdf_type=rdfvalue.RDFInteger,
        )
    )

    self.assertIsInstance(sample.test, rdfvalue.RDFInteger)

    # Can we assign an RDFValue instance?
    sample.test = rdfvalue.RDFInteger(5)

    self.assertEqual(sample.test, 5)

    # Check that bare values can be coerced.
    sample.test = 6
    self.assertIsInstance(sample.test, rdfvalue.RDFInteger)
    self.assertEqual(sample.test, 6)

    # Assign an enum.
    sample.registry_type = sample.RegistryType.REG_DWORD
    self.assertEqual(sample.registry_type, sample.RegistryType.REG_DWORD)

    sample.registry_type = rdf_client_fs.StatEntry.RegistryType.REG_SZ
    self.assertEqual(sample.registry_type, sample.RegistryType.REG_SZ)

    # We can also assign the string value.
    sample.registry_type = "REG_QWORD"
    self.assertEqual(sample.registry_type, sample.RegistryType.REG_QWORD)

    # Check that coercing works.
    sample.test = "10"
    self.assertEqual(sample.test, 10)

    # Assign an RDFValue which can not be coerced.
    self.assertRaises(
        type_info.TypeValueError,
        setattr,
        sample,
        "test",
        rdfvalue.RDFString("hello"),
    )

  def testComplexConstruction(self):
    """Test that we can construct RDFProtos with nested fields."""
    pathspec = rdf_paths.PathSpec(
        path="/foobar", pathtype=rdf_paths.PathSpec.PathType.TSK
    )
    sample = rdf_client_fs.StatEntry(pathspec=pathspec, st_size=5)

    self.assertEqual(sample.pathspec.path, "/foobar")
    self.assertEqual(sample.st_size, 5)

    self.assertRaises(AttributeError, rdf_client_fs.StatEntry, foobar=1)

  def testUnicodeSupport(self):
    pathspec = rdf_paths.PathSpec(
        path="/foobar", pathtype=rdf_paths.PathSpec.PathType.TSK
    )
    pathspec.path = "Grüezi"

    self.assertEqual(pathspec.path, "Grüezi")

  def testRepeatedFields(self):
    """Test handling of protobuf repeated fields."""
    sample = rdf_client_network.Interface()

    # Add an invalid type.
    self.assertRaises(type_info.TypeValueError, sample.addresses.Append, 2)

    # Add an rdfvalue by kwargs.
    sample.addresses.Append(human_readable_address="127.0.0.1")

    self.assertEqual(sample.addresses[0].human_readable_address, "127.0.0.1")
    self.assertLen(sample.addresses, 1)

    # Add an rdfvalue.
    sample.addresses.Append(
        rdf_client_network.NetworkAddress(human_readable_address="1.2.3.4")
    )

    self.assertLen(sample.addresses, 2)
    self.assertEqual(sample.addresses[1].human_readable_address, "1.2.3.4")

  def testEnums(self):
    """Check that enums are wrapped in a descriptor class."""
    sample = rdf_flows.GrrStatus()
    self.assertEqual(sample.status, rdf_flows.GrrStatus.ReturnedStatus.OK)


class EnumNamedValueTest(absltest.TestCase):

  def testInitialization(self):
    a = rdf_structs.EnumNamedValue(initializer=5, name="A", description="B")
    self.assertEqual(a.id, 5)
    self.assertEqual(a.name, "A")
    self.assertEqual(a.description, "B")

    b = rdf_structs.EnumNamedValue()
    self.assertEqual(b.id, 0)
    self.assertEqual(b.name, "0")
    self.assertIsNone(b.description)

  def testComparison(self):
    a = rdf_structs.EnumNamedValue(initializer=0, name="A")
    b = rdf_structs.EnumNamedValue(initializer=1, name="B")

    self.assertEqual(a, rdf_structs.EnumNamedValue(0, name="A"))
    self.assertNotEqual(a, b)
    self.assertLess(a, b)
    self.assertGreater(b, a)

    self.assertEqual(b, 1)
    self.assertEqual(b, "B")

  def testSerialization(self):
    # Current EnumNamedValue implementation only persists the ID.
    a = rdf_structs.EnumNamedValue(initializer=3)
    a2 = rdf_structs.EnumNamedValue.FromSerializedBytes(a.SerializeToBytes())

    self.assertEqual(a2.id, 3)
    self.assertEqual(a, a2)

    a2 = rdf_structs.EnumNamedValue.FromWireFormat(a.SerializeToWireFormat())
    self.assertEqual(a2.id, 3)
    self.assertEqual(a, a2)

  def testHumanReadable(self):
    a = rdf_structs.EnumNamedValue(initializer=5, name="A")
    self.assertEqual(str(a), "A")

    a2 = rdf_structs.EnumNamedValue.FromHumanReadable(str(a))
    self.assertEqual(a2.name, "A")
    self.assertEqual(a2, a)


class EnumContainerTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.enum_container = rdf_structs.EnumContainer(
        name="foo", values={"bar": 1, "baz": 2}
    )

  def testFromString(self):
    self.assertEqual(
        self.enum_container.FromString("bar"), self.enum_container.bar
    )

  def testFromString_invalidValue(self):
    with self.assertRaises(ValueError):
      self.enum_container.FromString("bax")


class NoDynamicTypeLookupTest(absltest.TestCase):

  class NoDynamicTypeLookupMessage(rdf_structs.RDFProtoStruct):

    protobuf = tests_pb2.NoDynamicTypeLookupMessage

  def testFieldAccessors(self):
    message = self.NoDynamicTypeLookupMessage()
    self.assertIsInstance(message.any, rdf_structs.AnyValue)

    message.any.type_url = "foo.bar.baz"
    message.any.value = b"quux"

    self.assertEqual(message.any.type_url, "foo.bar.baz")
    self.assertEqual(message.any.value, b"quux")

  def testSerializeAndDeserialize(self):
    message = self.NoDynamicTypeLookupMessage()
    message.any.type_url = "foo.bar.baz"
    message.any.value = b"quux"

    blob = message.SerializeToBytes()
    message = self.NoDynamicTypeLookupMessage.FromSerializedBytes(blob)

    self.assertEqual(message.any.type_url, "foo.bar.baz")
    self.assertEqual(message.any.value, b"quux")

  def testAsPrimitiveProto(self):
    message = self.NoDynamicTypeLookupMessage()
    message.any.type_url = "foo.bar.baz"
    message.any.value = b"quux"

    proto = message.AsPrimitiveProto()
    self.assertEqual(proto.any.type_url, "foo.bar.baz")
    self.assertEqual(proto.any.value, b"quux")


class AnyValueTest(absltest.TestCase):

  def testFromProto2(self):
    string = wrappers_pb2.StringValue(value="Lorem ipsum.")

    string_proto2_any = any_pb2.Any()
    string_proto2_any.Pack(string)

    string_rdf_any = rdf_structs.AnyValue.FromProto2(string_proto2_any)

    unpacked_string = wrappers_pb2.StringValue()
    unpacked_string.ParseFromString(string_rdf_any.value)

    self.assertEqual(unpacked_string, string)

  def testPack(self):
    user = rdf_client.User()
    user.username = "foo"

    proto = rdf_structs.AnyValue.Pack(user)
    self.assertIn("User", proto.type_url)

    deserialized = rdf_client.User.FromSerializedBytes(proto.value)
    self.assertEqual(deserialized.username, "foo")

  def testUnpack(self):
    user = rdf_client.User()
    user.username = "foo"

    proto = rdf_structs.AnyValue.Pack(user)

    unpacked = proto.Unpack(rdf_client.User)
    self.assertEqual(unpacked.username, "foo")

  def testUnpackIncompatibleTypeURL(self):
    user = rdf_client.User()
    user.username = "foo"

    proto = rdf_structs.AnyValue.Pack(user)
    proto.type_url = "type.googleapis.com/com.example.Foo"

    unpacked = proto.Unpack(rdf_client.User)
    self.assertEqual(unpacked.username, "foo")

  def testPackProto2(self):
    string = wrappers_pb2.StringValue(value="Lorem ipsum.")

    packed_string = rdf_structs.AnyValue.PackProto2(string)
    self.assertIn("StringValue", packed_string.type_url)

    unpacked_string = wrappers_pb2.StringValue()
    unpacked_string.ParseFromString(packed_string.value)

    self.assertEqual(unpacked_string, string)

  def testPackProto2RDFValue(self):
    user = rdf_client.User()
    user.username = "foo"

    packed_user = rdf_structs.AnyValue.PackProto2(user.AsPrimitiveProto())

    unpacked_user = packed_user.Unpack(rdf_client.User)
    self.assertEqual(unpacked_user, user)


class VarintTest(absltest.TestCase):
  """Tests the VarintEncode and VarintReader implementations."""

  _ENCODED_VARINTS = (
      (1, b"\x01"),
      (1 << 4, b"\x10"),
      (1 << 8, b"\x80\x02"),
      (1 << 12, b"\x80\x20"),
      (1 << 16, b"\x80\x80\x04"),
      (1 << 62, b"\x80\x80\x80\x80\x80\x80\x80\x80\x40"),
      (1 << 63, b"\x80\x80\x80\x80\x80\x80\x80\x80\x80\x01"),
  )

  def testCorrectlyEncodesIntegers(self):
    for val, b in self._ENCODED_VARINTS:
      with self.subTest(value=val):
        self.assertEqual(rdf_structs.VarintEncode(val), b)

  def testEncodesOverflowIntegerAsZero(self):
    self.assertEqual(rdf_structs.VarintEncode(1 << 66), b"\x00")

  def testCorrectlyReadsEncodedIntegers(self):
    for src_val, b in self._ENCODED_VARINTS:
      with self.subTest(buffer=b):
        self.assertEqual(rdf_structs.VarintReader(b, 0), (src_val, len(b)))

  def testReaderCanReadWhatEncoderHasEncoded(self):
    for v in [
        0,
        1,
        2,
        10,
        20,
        63,
        64,
        65,
        127,
        128,
        129,
        255,
        256,
        257,
        1 << 63 - 1,
        1 << 64 - 1,
    ]:
      with self.subTest(value=v):
        buf = rdf_structs.VarintEncode(v)
        r, p = rdf_structs.VarintReader(buf, 0)
        self.assertEqual(v, r)
        self.assertLen(buf, p)

  def testDecodingZeroBufferRaises(self):
    with self.assertRaisesRegex(
        ValueError, "Too many bytes when decoding varint"
    ):
      rdf_structs.VarintReader(b"", 0)

  def testDecodingValueLargerThan64BitReturnsTruncatedValue(self):
    self.assertEqual(
        rdf_structs.VarintReader(
            b"\x80\x80\x80\x80\x80\x80\x80\x80\x80\x02", 0
        ),
        (0, 10),
    )

  def testRaisesWhenBufferIsTooLong(self):
    with self.assertRaisesRegex(
        ValueError, "Too many bytes when decoding varint"
    ):
      rdf_structs.VarintReader(
          b"\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x01\x00\x00", 0
      )
    with self.assertRaisesRegex(
        ValueError, "Too many bytes when decoding varint"
    ):
      rdf_structs.VarintReader(
          b"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF", 0
      )

  def testNonCanonicalZeroDoesNotRaise(self):
    self.assertEqual(rdf_structs.VarintReader(b"\x80\x80\x80\x00", 0), (0, 4))


class SignedVarintTest(absltest.TestCase):
  """Test the VarintReader implementation."""

  _ENCODED_SIGNED_VARINTS = (
      (-1, b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01"),
      (-(1 << 4), b"\xf0\xff\xff\xff\xff\xff\xff\xff\xff\x01"),
      (-(1 << 8), b"\x80\xfe\xff\xff\xff\xff\xff\xff\xff\x01"),
      (-(1 << 12), b"\x80\xe0\xff\xff\xff\xff\xff\xff\xff\x01"),
      (-(1 << 16), b"\x80\x80\xfc\xff\xff\xff\xff\xff\xff\x01"),
      (-(1 << 62), b"\x80\x80\x80\x80\x80\x80\x80\x80\xc0\x01"),
      (-(1 << 63), b"\x80\x80\x80\x80\x80\x80\x80\x80\x80\x01"),
  )

  def testCorrectlyEncodesSignedIntegers(self):
    for val, b in self._ENCODED_SIGNED_VARINTS:
      with self.subTest(value=val):
        self.assertEqual(rdf_structs.SignedVarintEncode(val), b)

  def testEncodesOverflowIntegerAsZero(self):
    self.assertEqual(rdf_structs.SignedVarintEncode(-(1 << 66)), b"\x00")

  def testCorrectlyReadsEncodedSignedIntegers(self):
    for src_val, b in self._ENCODED_SIGNED_VARINTS:
      with self.subTest(buffer=b):
        self.assertEqual(
            rdf_structs.SignedVarintReader(b, 0), (src_val, len(b))
        )


class SplitBufferTest(absltest.TestCase):
  """Test the SplitBuffer function."""

  def testRaisesIfIndexOrLengthIncorrect(self):
    buf = b"\x00\x00"
    with self.assertRaises(ValueError):
      rdf_structs.SplitBuffer(buf, -1, 2)
    with self.assertRaises(ValueError):
      rdf_structs.SplitBuffer(buf, 0, -1)
    with self.assertRaises(ValueError):
      rdf_structs.SplitBuffer(buf, -1, -1)
    with self.assertRaises(ValueError):
      rdf_structs.SplitBuffer(buf, 3, 2)

  def testReturnsNothingOnEmptyBuffer(self):
    self.assertEqual(rdf_structs.SplitBuffer(b"", 0, 0), [])
    self.assertEqual(rdf_structs.SplitBuffer(b"\x00\x00", 2, 2), [])

  def testRaisesOnBrokenTag(self):
    with self.assertRaisesRegex(ValueError, "Broken tag encountered"):
      rdf_structs.SplitBuffer(b"\xff", 0, 1)

  def testRaisesOnBrokenVarintTag(self):
    # This used to trigger an infinite loop in the accelerated.c implementation.
    with self.assertRaisesRegex(ValueError, "Broken varint tag encountered"):
      rdf_structs.SplitBuffer(b"\x00", 0, 1)

    with self.assertRaisesRegex(ValueError, "Broken varint tag encountered"):
      rdf_structs.SplitBuffer(b"\x00\xff", 0, 2)

  def testCorrectlyProcessesVarintTag(self):
    self.assertEqual(
        rdf_structs.SplitBuffer("\x00\x01", 0, 2), [(b"\x00", b"", b"\x01")]
    )

  def testRaisesOnOversizedFixed64Tag(self):
    for l in range(8):
      with self.subTest(l=l):
        with self.assertRaisesRegex(
            ValueError, "Fixed64 tag exceeds available buffer"
        ):
          buf = b"\x01" + b"\x00" * l
          rdf_structs.SplitBuffer(buf, 0, len(buf))

  def testCorrectlyProcessesFixed64Tag(self):
    self.assertEqual(
        rdf_structs.SplitBuffer("\x01\x00\x01\x02\x03\x04\x05\0x6\0x7", 0, 9),
        [(b"\x01", b"", b"\x00\x01\x02\x03\x04\x05\x00x")],
    )

  def testRaisesOnOversizedFixed32Tag(self):
    for l in range(4):
      with self.subTest(l=l):
        with self.assertRaisesRegex(
            ValueError, "Fixed32 tag exceeds available buffer"
        ):
          buf = b"\x05" + b"\x00" * l
          rdf_structs.SplitBuffer(buf, 0, len(buf))

  def testCorrectlyProcessesFixed32Tag(self):
    self.assertEqual(
        rdf_structs.SplitBuffer("\x05\x00\x01\x02\x03", 0, 5),
        [(b"\x05", b"", b"\x00\x01\x02\x03")],
    )

  def testRaisesOnBrokenLengthDelimitedTag(self):
    with self.assertRaisesRegex(
        ValueError, "Broken length_delimited tag encountered"
    ):
      rdf_structs.SplitBuffer(b"\x02\xff", 0, 2)

  def testRaisesOnLengthDelimitedTagExceedingMaxInt(self):
    with self.assertRaisesRegex(ValueError, "Length delimited exceeds limits"):
      buf = b"\x0a\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01\x02\xff\xff\xff\xff\xff\xff\x27"
      rdf_structs.SplitBuffer(buf, 0, len(buf))

  def testRaisesOnOversizedLengthDelimitedTag(self):
    with self.assertRaisesRegex(
        ValueError, "Length tag exceeds available buffer"
    ):
      rdf_structs.SplitBuffer(b"\x02\x02", 0, 2)

  def testCorrectlyProcessesLengthDelimitedTag(self):
    self.assertEqual(
        rdf_structs.SplitBuffer("\x02\x02\x00\x01", 0, 4),
        [(b"\x02", b"\x02", b"\x00\x01")],
    )

  def testRaisesOnUnknownTag(self):
    with self.assertRaisesRegex(ValueError, "Unexpected Tag"):
      rdf_structs.SplitBuffer(b"\x03", 0, 1)


def main(argv):
  test_lib.main(argv)


if __name__ == "__main__":
  app.run(main)



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import plist as rdf_plist
from grr_response_proto import sysinfo_pb2


def ToProtoPlistBoolDictEntry(
    rdf: rdf_plist.PlistBoolDictEntry,
) -> sysinfo_pb2.PlistBoolDictEntry:
  return rdf.AsPrimitiveProto()


def ToRDFPlistBoolDictEntry(
    proto: sysinfo_pb2.PlistBoolDictEntry,
) -> rdf_plist.PlistBoolDictEntry:
  return rdf_plist.PlistBoolDictEntry.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoPlistStringDictEntry(
    rdf: rdf_plist.PlistStringDictEntry,
) -> sysinfo_pb2.PlistStringDictEntry:
  return rdf.AsPrimitiveProto()


def ToRDFPlistStringDictEntry(
    proto: sysinfo_pb2.PlistStringDictEntry,
) -> rdf_plist.PlistStringDictEntry:
  return rdf_plist.PlistStringDictEntry.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoLaunchdStartCalendarIntervalEntry(
    rdf: rdf_plist.LaunchdStartCalendarIntervalEntry,
) -> sysinfo_pb2.LaunchdStartCalendarIntervalEntry:
  return rdf.AsPrimitiveProto()


def ToRDFLaunchdStartCalendarIntervalEntry(
    proto: sysinfo_pb2.LaunchdStartCalendarIntervalEntry,
) -> rdf_plist.LaunchdStartCalendarIntervalEntry:
  return rdf_plist.LaunchdStartCalendarIntervalEntry.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoLaunchdKeepAlive(
    rdf: rdf_plist.LaunchdKeepAlive,
) -> sysinfo_pb2.LaunchdKeepAlive:
  return rdf.AsPrimitiveProto()


def ToRDFLaunchdKeepAlive(
    proto: sysinfo_pb2.LaunchdKeepAlive,
) -> rdf_plist.LaunchdKeepAlive:
  return rdf_plist.LaunchdKeepAlive.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoLaunchdPlist(
    rdf: rdf_plist.LaunchdPlist,
) -> sysinfo_pb2.LaunchdPlist:
  return rdf.AsPrimitiveProto()


def ToRDFLaunchdPlist(
    proto: sysinfo_pb2.LaunchdPlist,
) -> rdf_plist.LaunchdPlist:
  return rdf_plist.LaunchdPlist.FromSerializedBytes(proto.SerializeToString())



#!/usr/bin/env python
"""Configuration parameters for logging and error reporting subsystems."""


from grr_response_core.lib import config_lib
from grr_response_core.lib import type_info
from grr_response_core.lib.rdfvalues import standard as rdf_standard

config_lib.DEFINE_string(
    "Logging.domain", "localhost",
    "The email domain belonging to this installation. "
    "Leave blank to not restrict email to this domain")

config_lib.DEFINE_list(
    "Logging.engines", ["stderr"], "Enabled logging engines. Valid values are "
    "combinations of stderr,file,syslog,event_log.")

config_lib.DEFINE_bool(
    "Logging.verbose", False, help="If true log more verbosely.")

config_lib.DEFINE_string(
    "Logging.path",
    "",
    help="Path to log file directory.",
)

config_lib.DEFINE_string(
    "Logging.syslog_path",
    "/dev/log",
    help="Path to syslog socket. This can be a unix "
    "domain socket or in a UDP host:port notation.")

config_lib.DEFINE_string(
    "Logging.filename",
    "%(Logging.path)/GRRlog.txt",
    help="Filename of the grr log file.")

config_lib.DEFINE_option(
    type_info.RDFValueType(
        rdfclass=rdf_standard.DomainEmailAddress,
        name="Monitoring.alert_email",
        description="The email address to send events to.",
        default=None))

config_lib.DEFINE_option(
    type_info.RDFValueType(
        rdfclass=rdf_standard.DomainEmailAddress,
        name="Monitoring.emergency_access_email",
        description="The email address to notify in an emergency.",
        default="grr-emergency@localhost"))

config_lib.DEFINE_string("Monitoring.http_address", "::1",
                         "IP address for stats monitoring server.")

config_lib.DEFINE_integer("Monitoring.http_port", 0,
                          "Port for stats monitoring server.")

config_lib.DEFINE_integer(
    "Monitoring.http_port_max", None,
    "If set and Monitoring.http_port is in use, attempt "
    "to use ports between Monitoring.http_port and "
    "Monitoring.http_port_max.")



#!/usr/bin/env python
"""Common logic for objects used to track monitoring-related metrics.

GRR has a notion of statistic metrics. These metrics reflect GRR's state at
any given moment. There are three types of metrics available:

1. Counter. This is an integer metric that can only be incremented and never
decremented. Example: number of handled requests.
2. Gauge. Gauge metrics have a type, they can be either string, integer
or float. Gauge metric may have any value of its type.
3. Event. Event metrics are used to record events that take certain amount of
time. They're stored as latency distributions. Example: request latency.

Metrics may have fields. Fields are used in cases where you would generally
require a dynamically named metric. For example if you have requests coming
from http and rpc sources. You can defined 2 metrics: requests_count_http and
requests_count_rpc or define a single metric with a field "source".

Fields are essentially dimensions. For example,
if a metric "request_count" has no fields, it will be stored as a simple list
of data. I.e.:
request_count: 10 11 12 13 14

If "request_count" metric has a field "source" of type "str", then it will be
stored as a table. I.e. it will store different lists of values for different
values of the source field:
request_count[source=http]: 10 11 12 13 14
request_count[source=rpc]:  0  1  2  4  5

If "request_count" metric has a field "source" of type "str" and a field
"datacenter_index" of type "int", then it will be stored as a 3-dimensional
table. I.e.:
request_count[source=http,datacenter_index=0]: 10 11 12 13 14
request_count[source=rpc,datacenter_index=0]:  0  0  0  2  2
request_count[source=http,datacenter_index=1]: 33 34 45 88 99
request_count[source=rpc,datacenter_index=1]:  0  2  3  4  4
request_count[source=http,datacenter_index=2]: 22 33 44 55 66
request_count[source=rpc,datacenter_index=2]:  10 11 11 20 21

To summarize, for every combination of different field values, a separate row
of statistics data will be collected. Given that it's extremely important
to ensure that every field used in particular metric has a finite number of
possible values.
"""

import abc
import threading

from grr_response_core.lib import utils


class StatsCollector(metaclass=abc.ABCMeta):
  """Base implementation for a stats-collector.

  Stats-collectors are used to track monitoring-related metrics. By default,
  metrics are effectively single-dimensional. However, one or more named
  dimensions can be specified when defining a metric. Thereafter, the metric
  gets updated by specifying values for each of the dimensions/fields. For
  example, if a counter is initialized with the dimensions
    [("method_name", str), ("response_code", int)],
  an example of how it gets incremented would be:
    collector.IncrementCounter("counter_name", fields=["POST", 200])

  Attributes:
    lock: threading.RLock required by the utils.Synchronized decorator.
  """

  def __init__(self):
    self._metadata_dict = {}
    self.lock = threading.RLock()

  @utils.Synchronized
  def RegisterMetric(self, metadata):
    if metadata.varname in self._metadata_dict:
      raise ValueError("Duplicate metadata for metric %s." % metadata.varname)
    self._InitializeMetric(metadata)
    self._metadata_dict[metadata.varname] = metadata

  @abc.abstractmethod
  def _InitializeMetric(self, metadata):
    """Initializes a metric with the given metadata.

    Args:
      metadata: MetricMetadata for the metric.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def IncrementCounter(self, metric_name, delta=1, fields=None):
    """Increments a counter metric by a given delta.

    Args:
      metric_name: Name of the metric.
      delta: Delta by which the metric should be incremented.
      fields: Values for this metric's dimensions. Should only be provided if
        the metric was registered with dimensions.

    Raises:
      ValueError: If delta < 0.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def RecordEvent(self, metric_name, value, fields=None):
    """Records value corresponding to the given event metric.

    Args:
      metric_name: Name of the metric.
      value: Value to be recorded.
      fields: Values for this metric's dimensions. Should only be provided if
        the metric was registered with dimensions.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def SetGaugeValue(self, metric_name, value, fields=None):
    """Sets value of a given gauge metric.

    Args:
      metric_name: Name of the metric.
      value: New metric value.
      fields: Values for this metric's dimensions. Should only be provided if
        the metric was registered with dimensions.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def SetGaugeCallback(self, metric_name, callback, fields=None):
    """Attaches a callback to the given gauge metric.

    Args:
      metric_name: Name of the metric.
      callback: Zero argument function that is expected to provide the metric's
        current value (with corresponding field values) every time it is called.
      fields: Values for this metric's dimensions. Should only be provided if
        the metric was registered with dimensions. If provided, the callback
        will only be invoked when caller provides same field values.
    """
    raise NotImplementedError()

  def GetMetricMetadata(self, metric_name):
    """Returns the MetricMetadata for the given metric.

    Args:
      metric_name: Name of the metric.

    Returns:
      MetricMetadata object describing the metric.

    Raises:
      KeyError: if metric is not found.
    """
    return self._metadata_dict[metric_name]

  def GetAllMetricsMetadata(self):
    """Returns a dict mapping all metric names to their MetricMetadata."""
    return self._metadata_dict

  @abc.abstractmethod
  def GetMetricFields(self, metric_name):
    """Returns all field values for the given metric.

    Args:
      metric_name: Name of the metric.

    Returns:
      A list of tuples containing field values that were used for the specified
      metric. For example, if there's a counter metric registered with fields
      ("renderer_type", int) and there were 2 calls to Increment:
        Increment("renderer_type", fields=[1])
        Increment("renderer_type", fields=[2]),
      then GetMetricFields("renderer_type") will return [(1,), (2,)].
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def GetMetricValue(self, metric_name, fields=None):
    """Returns the value of a given metric for given field values.

    Args:
      metric_name: Name of the metric.
      fields: List of values for this metric's dimensions. Should only be
        provided if the metric was registered without any dimensions.

    Returns:
      int for a counter metric.
      int, float, str (depending on the type used to register
        the metric) for a gauge metric.
      Distribution-compatible object for event metric. Distribution-compatible
      means "with an API matching the API of the Distribution object".
    """
    raise NotImplementedError()



#!/usr/bin/env python
"""Metric implementations to collect statistics."""

import abc

from grr_response_core.lib.rdfvalues import stats as rdf_stats
from grr_response_core.stats import stats_collector_instance
from grr_response_core.stats import stats_utils


class AbstractMetric(metaclass=abc.ABCMeta):
  """An abstract metric with a name, fields, and values.

  Refer to default_stats_collector._Metric and DefaultStatsCollector to
  see how StatsCollector handles the field definitions and values.

  Attributes:
    name: string containing the global metric name.
  """

  def __init__(self, metadata):
    """Initializes a new metric and registers it with the StatsCollector."""
    self.name = metadata.varname
    stats_collector_instance.RegisterMetric(metadata)

  def GetValue(self, fields=None):
    """Returns the value of a given metric for given field values."""
    return stats_collector_instance.Get().GetMetricValue(
        self.name, fields=fields
    )

  def GetFields(self):
    """Returns all field values for the given metric."""
    return stats_collector_instance.Get().GetMetricFields(self.name)


class Counter(AbstractMetric):
  """A Counter metric that can be incremented.

  Refer to default_stats_collector._CounterMetric and DefaultStatsCollector to
  see how StatsCollector handles the field definitions and values.
  """

  def __init__(self, name, fields=(), docstring=None, units=None):
    """Initializes a Counter metric and registers it with the StatsCollector."""
    super().__init__(
        rdf_stats.MetricMetadata(
            varname=name,
            metric_type=rdf_stats.MetricMetadata.MetricType.COUNTER,
            value_type=rdf_stats.MetricMetadata.ValueType.INT,
            fields_defs=stats_utils.FieldDefinitionProtosFromTuples(fields),
            docstring=docstring,
            units=units,
        )
    )

  def Increment(self, delta=1, fields=None):
    """Increments a counter metric by a given delta."""
    stats_collector_instance.Get().IncrementCounter(
        self.name, delta, fields=fields
    )

  def Counted(self, fields=None):
    """Returns a decorator that counts function calls."""
    return stats_utils.Counted(self, fields=fields)

  def SuccessesCounted(self, fields=None):
    """Returns a decorator that counts calls that don't raise an exception."""
    return stats_utils.SuccessesCounted(self, fields=fields)

  def ErrorsCounted(self, fields=None):
    """Returns a decorator that counts calls that raise an exception."""
    return stats_utils.ErrorsCounted(self, fields=fields)


class Gauge(AbstractMetric):
  """A Gauge metric that can be set to a value.

  Refer to default_stats_collector._GaugeMetric and DefaultStatsCollector to
  see how StatsCollector handles the field definitions and values.
  """

  def __init__(self, name, value_type, fields=(), docstring=None, units=None):
    """Initializes a Gauge metric and registers it with the StatsCollector."""
    super().__init__(
        rdf_stats.MetricMetadata(
            varname=name,
            metric_type=rdf_stats.MetricMetadata.MetricType.GAUGE,
            value_type=stats_utils.MetricValueTypeFromPythonType(value_type),
            fields_defs=stats_utils.FieldDefinitionProtosFromTuples(fields),
            docstring=docstring,
            units=units,
        )
    )

  def SetValue(self, value, fields=None):
    """Sets value of a given gauge metric."""
    stats_collector_instance.Get().SetGaugeValue(
        self.name, value, fields=fields
    )

  def SetCallback(self, callback, fields=None):
    """Attaches a callback to the given gauge metric."""
    stats_collector_instance.Get().SetGaugeCallback(
        self.name, callback, fields=fields
    )


class Event(AbstractMetric):
  """An Event metric that records timings of events.

  Refer to default_stats_collector._EventMetric and DefaultStatsCollector to
  see how StatsCollector handles the field definitions and values.
  """

  def __init__(self, name, bins=(), fields=(), docstring=None, units=None):
    """Initializes an Event metric and registers it with the StatsCollector."""
    super().__init__(
        rdf_stats.MetricMetadata(
            varname=name,
            bins=bins,
            metric_type=rdf_stats.MetricMetadata.MetricType.EVENT,
            value_type=rdf_stats.MetricMetadata.ValueType.DISTRIBUTION,
            fields_defs=stats_utils.FieldDefinitionProtosFromTuples(fields),
            docstring=docstring,
            units=units,
        )
    )

  def RecordEvent(self, value, fields=None):
    """Records value corresponding to the given event metric."""
    stats_collector_instance.Get().RecordEvent(self.name, value, fields=fields)

  def Timed(self, fields=None):
    """Returns a decorator that records timing metrics for function calls."""
    return stats_utils.Timed(self, fields=fields)



#!/usr/bin/env python
import contextlib
import functools
import io

from absl.testing import absltest

from grr_response_core.lib.util import context
from grr_response_core.lib.util import temp


class MultiContextTest(absltest.TestCase):

  def testEmpty(self):
    with context.MultiContext([]) as values:
      self.assertEqual(values, [])

  def testWithNulls(self):
    foo = contextlib.nullcontext("foo")
    bar = contextlib.nullcontext("bar")
    baz = contextlib.nullcontext("baz")

    with context.MultiContext([foo, bar, baz]) as names:
      self.assertEqual(names, ["foo", "bar", "baz"])

  def testWithFiles(self):
    foo = temp.AutoTempFilePath(suffix="foo")
    bar = temp.AutoTempFilePath(suffix="bar")
    baz = temp.AutoTempFilePath(suffix="baz")

    with context.MultiContext([foo, bar, baz]) as filepaths:
      self.assertLen(filepaths, 3)
      self.assertEndsWith(filepaths[0], "foo")
      self.assertEndsWith(filepaths[1], "bar")
      self.assertEndsWith(filepaths[2], "baz")

      wbopen = functools.partial(io.open, mode="wb")
      with context.MultiContext(list(map(wbopen, filepaths))) as filedescs:
        self.assertLen(filedescs, 3)
        filedescs[0].write(b"FOO")
        filedescs[1].write(b"BAR")
        filedescs[2].write(b"BAZ")

      # At this point all three files should be correctly written, closed and
      # ready for reading.

      rbopen = functools.partial(io.open, mode="rb")
      with context.MultiContext(list(map(rbopen, filepaths))) as filedescs:
        self.assertLen(filedescs, 3)
        self.assertEqual(filedescs[0].read(), b"FOO")
        self.assertEqual(filedescs[1].read(), b"BAR")
        self.assertEqual(filedescs[2].read(), b"BAZ")


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import cronjobs as rdf_cronjobs
from grr_response_proto import sysinfo_pb2


def ToProtoCronTabEntry(
    rdf: rdf_cronjobs.CronTabEntry,
) -> sysinfo_pb2.CronTabEntry:
  return rdf.AsPrimitiveProto()


def ToRDFCronTabEntry(
    proto: sysinfo_pb2.CronTabEntry,
) -> rdf_cronjobs.CronTabEntry:
  return rdf_cronjobs.CronTabEntry.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoCronTabFile(
    rdf: rdf_cronjobs.CronTabFile,
) -> sysinfo_pb2.CronTabFile:
  return rdf.AsPrimitiveProto()


def ToRDFCronTabFile(
    proto: sysinfo_pb2.CronTabFile,
) -> rdf_cronjobs.CronTabFile:
  return rdf_cronjobs.CronTabFile.FromSerializedBytes(proto.SerializeToString())



#!/usr/bin/env python
import ctypes
import os
import platform
import stat
import time

from absl.testing import absltest

from grr_response_core.lib.util import statx
from grr_response_core.lib.util import temp
from grr.test_lib import time as time_test_lib


class GetTest(absltest.TestCase):

  def testNlink(self):
    with temp.AutoTempDirPath(remove_non_empty=True) as tempdir:
      target = os.path.join(tempdir, "foo")
      link_1 = os.path.join(tempdir, "bar")
      link_2 = os.path.join(tempdir, "baz")

      with open(target, mode="wb"):
        pass

      os.link(target, link_1)
      os.link(target, link_2)

      result = statx.Get(target.encode("utf-8"))
      self.assertEqual(result.nlink, 3)

  @absltest.skipIf(platform.system() == "Windows", "UID not available")
  def testUid(self):
    with temp.AutoTempFilePath() as tempfile:
      result = statx.Get(tempfile.encode("utf-8"))
      self.assertEqual(result.uid, os.getuid())

  @absltest.skipIf(platform.system() == "Windows", "GID not available")
  def testGid(self):
    with temp.AutoTempFilePath() as tempfile:
      result = statx.Get(tempfile.encode("utf-8"))
      self.assertEqual(result.gid, os.getgid())

  def testModeReg(self):
    with temp.AutoTempFilePath() as tempfile:
      result = statx.Get(tempfile.encode("utf-8"))
      self.assertTrue(stat.S_ISREG(result.mode))

  def testModeDir(self):
    with temp.AutoTempDirPath() as tempdir:
      result = statx.Get(tempdir.encode("utf-8"))
      self.assertTrue(stat.S_ISDIR(result.mode))

  def testModeLink(self):
    with temp.AutoTempDirPath(remove_non_empty=True) as tempdir:
      target = os.path.join(tempdir, "foo")
      link = os.path.join(tempdir, "bar")

      with open(target, mode="wb"):
        pass

      os.symlink(target, link)

      result = statx.Get(link.encode("utf-8"))
      self.assertTrue(stat.S_ISLNK(result.mode))

  def testIno(self):
    with temp.AutoTempFilePath() as tempfile:
      result = statx.Get(tempfile.encode("utf-8"))
      self.assertNotEqual(result.ino, 0)

  def testSize(self):
    with temp.AutoTempFilePath() as tempfile:
      with open(tempfile, mode="wb") as tempfile_handle:
        tempfile_handle.write(b"A" * 42)

      result = statx.Get(tempfile.encode("utf-8"))
      self.assertEqual(result.size, 42)

  def testDev(self):
    with temp.AutoTempFilePath() as tempfile:
      result = statx.Get(tempfile.encode("utf-8"))
      self.assertEqual(result.dev, os.lstat(tempfile).st_dev)

  @absltest.skipIf(platform.system() == "Windows", "`rdev` unavailable")
  def testRdev(self):
    with temp.AutoTempFilePath() as tempfile:
      result = statx.Get(tempfile.encode("utf-8"))
      self.assertEqual(result.rdev, os.lstat(tempfile).st_rdev)

  def testBaseTime(self):
    with temp.AutoTempFilePath() as tempfile:
      result = statx.Get(tempfile.encode("utf-8"))

      self.assertGreater(result.atime_ns, 0)
      self.assertGreater(result.mtime_ns, 0)
      self.assertGreater(result.ctime_ns, 0)

      time_test_lib.Step()
      self.assertLess(result.atime_ns, time.time_ns())
      self.assertLess(result.mtime_ns, time.time_ns())
      self.assertLess(result.ctime_ns, time.time_ns())

  def testBirthTime(self):
    # On Linux, file birth time is collected only if `statx` is available.
    if platform.system() == "Linux":
      try:
        ctypes.CDLL("libc.so.6").statx
      except AttributeError:
        raise absltest.SkipTest("`statx` not available")

    with temp.AutoTempFilePath() as tempfile:
      result = statx.Get(tempfile.encode("utf-8"))

      self.assertGreater(result.btime_ns, 0)
      time_test_lib.Step()
      self.assertLess(result.btime_ns, time.time_ns())

  def testRaisesExceptionOnError(self):
    with temp.AutoTempDirPath() as tempdir:
      with self.assertRaises(OSError):
        statx.Get(os.path.join(tempdir, "non-existing-file").encode("utf-8"))


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""Tests for config_lib classes."""

import io
import ntpath
import os
import stat
from unittest import mock

from absl import app
from absl.testing import absltest
from absl.testing import flagsaver

from grr_response_core import config
from grr_response_core.lib import config_lib
from grr_response_core.lib import config_parser
from grr_response_core.lib import package
from grr_response_core.lib import rdfvalue
from grr_response_core.lib import type_info
from grr_response_core.lib import utils
from grr_response_core.lib.rdfvalues import crypto as rdf_crypto
from grr_response_core.lib.rdfvalues import file_finder as rdf_file_finder
from grr_response_core.lib.rdfvalues import paths as rdf_paths
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_core.lib.util import temp
from grr.test_lib import test_lib


class YamlConfigTest(absltest.TestCase):
  """Test the Yaml config file support."""

  @flagsaver.flagsaver(disallow_missing_config_definitions=True)
  def testParsing(self):
    conf = config_lib.GrrConfigManager()

    conf.DEFINE_list("Section1.test_list", ["a", "b"], "A test integer.")
    conf.DEFINE_integer("Section1.test", 0, "An integer")
    conf.DEFINE_integer("Section1.test2", 0, "An integer")
    self.assertRaises(
        config_lib.MissingConfigDefinitionError,
        conf.Initialize,
        parser=config_parser.YamlConfigFileParser,
        data="""
                      Section2.test: 2
                      """,
    )

    conf.DEFINE_string("Section2.test", "", "A string")
    conf.DEFINE_context("Client Context")
    conf.DEFINE_context("Windows Context")
    conf.Initialize(
        parser=config_parser.YamlConfigFileParser,
        data="""

# Configuration options can be written as long hand, dot separated parameters.
Section1.test: 2
Section1.test_list: x,y
Section2.test: 3%(Section1.test)

Client Context:
  Section1.test: 6
  Section1.test2: 1

  Windows Context:
    Section1.test: 10

Windows Context:
  Section1.test: 5
  Section1.test2: 2

""",
    )

    self.assertEqual(conf["Section1.test"], 2)

    # Test interpolation works.
    self.assertEqual(conf["Section2.test"], "32")
    self.assertEqual(conf["Section1.test_list"], ["x", "y"])

    self.assertEqual(
        conf.Get(
            "Section1.test_list", context=["Client Context", "Windows Context"]
        ),
        ["x", "y"],
    )

    # Test that contexts affect option selection.
    self.assertEqual(conf.Get("Section1.test", context=["Client Context"]), 6)

    self.assertEqual(conf.Get("Section1.test", context=["Windows Context"]), 5)

    context = ["Client Context", "Windows Context"]
    self.assertEqual(conf.Get("Section1.test", context=context), 10)

    context = ["Windows Context", "Client Context"]
    # Order of the context parameters should not matter.
    self.assertEqual(conf.Get("Section1.test", context=context), 10)

  def testConflictingContexts(self):
    """Test that conflicting contexts are resolved by precedence."""
    conf = config_lib.GrrConfigManager()

    conf.DEFINE_integer("Section1.test", 0, "An integer")
    conf.DEFINE_context("Client Context")
    conf.DEFINE_context("Platform:Windows")
    conf.DEFINE_context("Extra Context")
    conf.Initialize(
        parser=config_parser.YamlConfigFileParser,
        data="""

Section1.test: 2

Client Context:
  Section1.test: 6

Platform:Windows:
  Section1.test: 10

Extra Context:
  Section1.test: 15
""",
    )

    # Without contexts.
    self.assertEqual(conf.Get("Section1.test"), 2)

    # When running in the client context only.
    self.assertEqual(conf.Get("Section1.test", context=["Client Context"]), 6)

    # Later defined contexts (i.e. with later calls to AddContext()) are
    # stronger than earlier contexts. For example, contexts set the command line
    # --context option are stronger than contexts set by the running binary,
    # since they are added last.
    self.assertEqual(
        conf.Get(
            "Section1.test", context=["Client Context", "Platform:Windows"]
        ),
        10,
    )

    self.assertEqual(
        conf.Get(
            "Section1.test", context=["Platform:Windows", "Client Context"]
        ),
        6,
    )

  def testRemoveContext(self):
    """Test that conflicting contexts are resolved by precedence."""
    conf = config_lib.GrrConfigManager()

    conf.DEFINE_integer("Section1.test", 0, "An integer")
    conf.DEFINE_integer("Section1.test2", 9, "An integer")
    conf.DEFINE_context("Client Context")
    conf.DEFINE_context("Platform:Windows")
    conf.DEFINE_context("Extra Context")
    conf.Initialize(
        parser=config_parser.YamlConfigFileParser,
        data="""

Section1.test: 2

Client Context:
  Section1.test: 6
  Section1.test2: 8

Platform:Windows:
  Section1.test: 10

Extra Context:
  Section1.test: 15
""",
    )

    # Should be defaults, no contexts added
    self.assertEqual(conf.Get("Section1.test"), 2)
    self.assertEqual(conf.Get("Section1.test2"), 9)

    # Now with Client Context
    conf.AddContext("Client Context")
    self.assertEqual(conf.Get("Section1.test"), 6)
    self.assertEqual(conf.Get("Section1.test2"), 8)

    # Should be back to defaults
    conf.RemoveContext("Client Context")
    self.assertEqual(conf.Get("Section1.test"), 2)
    self.assertEqual(conf.Get("Section1.test2"), 9)

    # Now with Windows Context, test2 is still default
    conf.AddContext("Platform:Windows")
    self.assertEqual(conf.Get("Section1.test"), 10)
    self.assertEqual(conf.Get("Section1.test2"), 9)

    # Should be back to defaults
    conf.RemoveContext("Platform:Windows")
    self.assertEqual(conf.Get("Section1.test"), 2)
    self.assertEqual(conf.Get("Section1.test2"), 9)

  def testContextApplied(self):
    conf = config_lib.GrrConfigManager()

    conf.DEFINE_integer("Section1.test", 0, "An integer")
    conf.DEFINE_context("Client Context")
    conf.DEFINE_context("Unused Context")
    conf.Initialize(
        parser=config_parser.YamlConfigFileParser,
        data="""
Client Context:
  Section1.test: 6
""",
    )

    # Should be defaults, no contexts added
    self.assertFalse(conf.ContextApplied("Client Context"))
    self.assertFalse(conf.ContextApplied("Unused Context"))

    conf.AddContext("Client Context")
    self.assertTrue(conf.ContextApplied("Client Context"))
    self.assertFalse(conf.ContextApplied("Unused Context"))

  def testBackslashes(self):
    conf = config_lib.GrrConfigManager()

    conf.DEFINE_string("Section1.parameter", "", "A test.")
    conf.DEFINE_string("Section1.parameter2", "", "A test.")
    conf.DEFINE_string("Section1.parameter3", "", "A test.")

    conf.Initialize(
        parser=config_parser.YamlConfigFileParser,
        data=r"""

Section1.parameter: |
   a\\b\\c\\d

Section1.parameter2: |
   %(parameter)\\e

Section1.parameter3: |
   \%(a\\b\\c\\d\)
""",
    )

    self.assertEqual(conf.Get("Section1.parameter"), "a\\b\\c\\d")
    self.assertEqual(conf.Get("Section1.parameter2"), "a\\b\\c\\d\\e")
    self.assertEqual(conf.Get("Section1.parameter3"), "%(a\\b\\c\\d)")

  def testSemanticValueType(self):
    conf = config_lib.GrrConfigManager()
    conf.DEFINE_semantic_value(
        rdfvalue.DurationSeconds, "Section1.foobar", None, "Sample help."
    )
    conf.Initialize(
        parser=config_parser.YamlConfigFileParser,
        data="""
Section1.foobar: 6d
""",
    )

    value = conf.Get("Section1.foobar")
    self.assertIsInstance(value, rdfvalue.DurationSeconds)
    self.assertEqual(value, rdfvalue.Duration.From(6, rdfvalue.DAYS))

  def testSemanticStructType(self):
    conf = config_lib.GrrConfigManager()

    conf.DEFINE_semantic_struct(
        rdf_file_finder.FileFinderArgs, "Section1.foobar", [], "Sample help."
    )
    conf.Initialize(
        parser=config_parser.YamlConfigFileParser,
        data="""
Section1.foobar:
  paths:
    - "a/b"
    - "b/c"
  pathtype: "TSK"
""",
    )

    values = conf.Get("Section1.foobar")
    self.assertIsInstance(values, rdf_file_finder.FileFinderArgs)
    self.assertEqual(values.paths, ["a/b", "b/c"])
    self.assertEqual(values.pathtype, "TSK")

  def testSemanticEnum(self):
    conf = config_lib.GrrConfigManager()

    conf.DEFINE_semantic_enum(
        enum_container=rdf_paths.PathSpec.PathType,
        name="Foo.Bar",
        default=rdf_paths.PathSpec.PathType.TSK,
    )
    conf.Initialize(
        parser=config_parser.YamlConfigFileParser, data="Foo.Bar: NTFS"
    )

    value = conf.Get("Foo.Bar")
    self.assertIsInstance(value, rdf_structs.EnumNamedValue)
    self.assertEqual(value, "NTFS")
    self.assertEqual(value.id, 5)

  def testSemanticEnum_defaultValue(self):
    conf = config_lib.GrrConfigManager()

    conf.DEFINE_semantic_enum(
        enum_container=rdf_paths.PathSpec.PathType,
        name="Foo.Bar",
        default=rdf_paths.PathSpec.PathType.TSK,
    )
    conf.Initialize(parser=config_parser.YamlConfigFileParser, data="")

    value = conf.Get("Foo.Bar")
    self.assertIsInstance(value, rdf_structs.EnumNamedValue)
    self.assertEqual(value, "TSK")

  def testSemanticEnum_invalidValue(self):
    conf = config_lib.GrrConfigManager()

    conf.DEFINE_semantic_enum(
        enum_container=rdf_paths.PathSpec.PathType,
        name="Foo.Bar",
        default=rdf_paths.PathSpec.PathType.TSK,
    )
    conf.Initialize(
        parser=config_parser.YamlConfigFileParser, data="Foo.Bar: Invalid"
    )

    with self.assertRaises(ValueError):
      conf.Get("Foo.Bar")


class ConfigLibTest(test_lib.GRRBaseTest):
  """Tests for config functionality."""

  def testInit(self):
    """Testing initialization of a ConfigManager."""
    conf = config_lib.GrrConfigManager()
    conf.DEFINE_string("MemoryDriver.device_path", "Default Value", "Help")
    conf.DEFINE_context("Platform:Windows")
    conf.DEFINE_context("Client Context")
    conf.DEFINE_context("Platform:Linux")

    data = r"""
Platform:Linux:
    MemoryDriver.device_path: /dev/pmem

Platform:Windows:
    MemoryDriver.device_path: \\\\.\\pmem
"""

    conf.Initialize(parser=config_parser.YamlConfigFileParser, data=data)

    # Check that the linux client have a different value from the windows
    # client.
    self.assertEqual(
        conf.Get(
            "MemoryDriver.device_path",
            context=("Client Context", "Platform:Linux"),
        ),
        "/dev/pmem",
    )

    self.assertEqual(
        conf.Get(
            "MemoryDriver.device_path",
            context=("Client Context", "Platform:Windows"),
        ),
        r"\\.\pmem",
    )

  def testSet(self):
    """Test setting options."""
    # Test access methods.
    conf = config_lib.GrrConfigManager()
    conf.DEFINE_string("NewSection1.new_option1", "Default Value", "Help")
    conf.initialized = True

    conf.Set("NewSection1.new_option1", "New Value1")

    self.assertEqual(conf["NewSection1.new_option1"], "New Value1")

  def testSave(self):
    """Save the config and ensure it still works."""
    conf = config_lib.GrrConfigManager()
    config_file = os.path.join(self.temp_dir, "writeback.yaml")
    conf.SetWriteBack(config_file)
    conf.DEFINE_string("NewSection1.new_option1", "Default Value", "Help")

    conf.Set("NewSection1.new_option1", "New Value1")

    conf.Write()

    new_conf = config_lib.GrrConfigManager()
    new_conf.DEFINE_string("NewSection1.new_option1", "Default Value", "Help")

    new_conf.Initialize(filename=config_file)

    self.assertEqual(new_conf["NewSection1.new_option1"], "New Value1")

  def testQuotes(self):
    conf = config_lib.GrrConfigManager()
    conf.DEFINE_string(name="foo.bar", default='"baz"', help="Bar.")
    conf.DEFINE_string(name="foo.quux", default='"%(foo.bar)"', help="Quux.")

    conf.Initialize(data="")

    self.assertEqual(conf["foo.bar"], '"baz"')
    self.assertEqual(conf["foo.quux"], '""baz""')

  def testWritebackQuotes(self):

    def Config():
      conf = config_lib.GrrConfigManager()
      conf.DEFINE_string(name="foo.bar", default="", help="Bar.")
      conf.DEFINE_string(name="foo.baz", default='"%(foo.bar)"', help="Baz.")
      return conf

    with temp.AutoTempFilePath(suffix=".yaml") as confpath:
      writeback_conf = Config()
      writeback_conf.SetWriteBack(confpath)
      writeback_conf.Set("foo.bar", '"quux"')
      writeback_conf.Write()

      loaded_conf = Config()
      loaded_conf.Initialize(filename=confpath)

    self.assertEqual(loaded_conf["foo.bar"], '"quux"')
    self.assertEqual(loaded_conf["foo.baz"], '""quux""')

  def _SetupConfig(self, value):
    conf = config_lib.GrrConfigManager()
    config_file = os.path.join(self.temp_dir, "config.yaml")
    with io.open(config_file, "w") as fd:
      fd.write("Section1.option1: %s" % value)
    conf.DEFINE_string("Section1.option1", "Default Value", "Help")
    conf.Initialize(filename=config_file)
    return conf

  def testPersist(self):
    writeback_file = os.path.join(self.temp_dir, "writeback.yaml")

    conf = self._SetupConfig("Value1")
    conf.SetWriteBack(writeback_file)

    self.assertEqual(conf["Section1.option1"], "Value1")

    conf.Persist("Section1.option1")

    conf = self._SetupConfig("Value2")
    # This should give the persisted value back.
    conf.SetWriteBack(writeback_file)

    self.assertEqual(conf["Section1.option1"], "Value1")

    # Now overwrite the writeback from the config ("Value2").
    conf.Persist("Section1.option1")

    conf = self._SetupConfig("Value3")
    conf.SetWriteBack(writeback_file)

    self.assertEqual(conf["Section1.option1"], "Value2")

    # This new config has the same value as the current writeback file.
    conf = self._SetupConfig("Value2")
    conf.SetWriteBack(writeback_file)

    self.assertEqual(conf["Section1.option1"], "Value2")

    def DontCall():
      raise NotImplementedError("Write was called!")

    # If the value in config and writeback are the same, nothing is written.
    with mock.patch.object(conf, "Write", DontCall):
      conf.Persist("Section1.option1")

  def testPersistDoesntOverwriteCustomOptions(self):
    conf = config_lib.GrrConfigManager()
    writeback_file = os.path.join(self.temp_dir, "writeback.yaml")
    conf.SetWriteBack(writeback_file)

    conf.DEFINE_string("Section.option", "Default Value", "Help")
    conf.Set("Section.option", "custom")
    conf.Write()

    new_conf = config_lib.GrrConfigManager()
    new_conf.DEFINE_string("Section.option", "Default Value", "Help")
    new_config_file = os.path.join(self.temp_dir, "config.yaml")
    new_conf.Initialize(filename=new_config_file)
    new_conf.SetWriteBack(writeback_file)
    new_conf.Write()

    # At this point, the writeback file has a custom setting for
    # "Section.option" but new_conf has nothing set.
    with io.open(writeback_file) as fd:
      self.assertEqual(fd.read(), "Section.option: custom\n")

    # Calling persist does not change the custom value.
    new_conf.Persist("Section.option")

    with io.open(writeback_file) as fd:
      self.assertEqual(fd.read(), "Section.option: custom\n")

  def testFileFilters(self):
    filename = os.path.join(self.temp_dir, "f.txt")
    content = "testcontent"
    with io.open(filename, "w") as fd:
      fd.write(content)

    conf = config_lib.GrrConfigManager()
    conf.DEFINE_string("Valid.file", "%%(%s|file)" % filename, "test")
    conf.DEFINE_string(
        "Valid.optionalfile", "%%(%s|optionalfile)" % filename, "test"
    )
    conf.DEFINE_string("Invalid.file", "%(notafile|file)", "test")
    conf.DEFINE_string(
        "Invalid.optionalfile", "%(notafile|optionalfile)", "test"
    )

    conf.Initialize(data="")

    self.assertEqual(conf["Valid.file"], content)
    self.assertEqual(conf["Valid.optionalfile"], content)

    with self.assertRaises(config_lib.FilterError):
      conf["Invalid.file"]  # pylint: disable=pointless-statement

    self.assertEqual(conf["Invalid.optionalfile"], "")

  def testErrorDetection(self):
    """Check that invalid config files are detected immediately."""
    test_conf = """
[Section1]
test = val2"""
    conf = config_lib.GrrConfigManager()
    # Define test as an integer.
    conf.DEFINE_integer("Section1.test", 54, "A test integer.")

    conf.Initialize(data=test_conf)

    # This should raise since the config file is incorrect.
    errors = conf.Validate("Section1")
    self.assertIn(
        "Invalid value val2 for Integer", str(errors["Section1.test"])
    )

  def testCopyConfig(self):
    """Check we can copy a config and use it without affecting the old one."""
    conf = config.CONFIG.CopyConfig()
    conf.initialized = False
    conf.DEFINE_string("NewSection1.new_option1", "Default Value", "Help")
    conf.Set("NewSection1.new_option1", "New Value1")
    conf.initialized = True
    conf.Write()
    self.assertEqual(conf.Get("NewSection1.new_option1"), "New Value1")
    self.assertEqual(config.CONFIG.Get("NewSection1.new_option1", None), None)

  def testKeyConfigOptions(self):
    """Check that keys get read correctly from the config."""
    # Clone a test config object from the global config so it knows about Client
    # options.
    conf = config.CONFIG.MakeNewConfig()
    conf.Initialize(data="""
[Client]
private_key = -----BEGIN RSA PRIVATE KEY-----
    MIIEowIBAAKCAQEAzbuwYnKTTleU2F4zu1gI/BolzR74470j6wK7QrQp5b5Qkmfe
    vCdX540Ax3mFt6isChrhIXy4LmTvM5SeiCvs6su8ro7ZYqtbZjpoBWorESy5VAnn
    uiuwiAvI7IOPYbSu8e+mHQa893hLYv9q0kQe1CtlvQsKir8aQ0YC63UWhrcMcMYC
    jUaRkWo8JUWoCW/GuZjbuFF0ZZsA13cVSOhFg1dxvX4/+FXlnyZQbu2Aex4NU9NG
    Yetl+SP10Uq/c92wZ34ENU3rH8Kw8Tar/dkug9+hgrcgt7JRECXmho+WFrKM1H6z
    1CNe6uEALMfogNl9Mm2jCgavl1wbV92pIKT1pQIDAQABAoIBAAz6jaCm+T+HO2rg
    myizClK7JI4u6W/Wj1+Z+xh0wgg4wXT8VrBt1qK912Io0AR7gH+dB/2t6aGgTk3Q
    Tvr+CGgcj8E/CPGWUqxdkWaMDCj9IQ5El148N4VcTEyNanwfuFdSHBuMgLB89zEz
    i03grHA2nP24Jq4E6yIk1nYmW2mGoZSIrrSDG/VQj4NipeGxbwJsOGdFmzYlefk9
    lwC7Db9v2ocdytxlH7lxhVqEnOokLXT7VO/EPTyF3CcObZx2JG2P4d6KILcsRePg
    +xMTp0wmq4S3Hoiwi2+RZNH7aHn5igvcNbx0+Eu9CIKINuNwmw3c/2bAz6u48FUY
    8aUmQkECgYEA7cD3E8TqUTxQxCrcuHKVzbS5ThLvfSbIo7hctz2b9zshomRaeh1m
    qg3p5shQ0+p029H26mu8oi5e130QCmzbnhWoZZbVAY4SEe6+5SAF1d1Xhwn8jTZ5
    Vqo5B7ILKVvxHWkgdYFV9yi6c23/kSzFDZXW4QFt3eZgPv8wLxWU3uUCgYEA3YWh
    VcX2UDrQR8r4Qy9YHuQ85vW6O5QKvSz//Ckw9Z6TfFOf5DRmAHyt2IpnmmbeuhDD
    SsnXv1fsqJWcbl+/nP2cNuIhLXnUwebtTW3Qo9PSi7APhOYEYAbGuDiJ8swPof42
    5JjnIJ70dQv4wF1qHfu+ya+LPNnRZW2gF6Hbj8ECgYB+/Yu7Xnl9rIbDUNWWG3YS
    as5zej+7DEUs1aOIKHsvAcGEWK/O+/dDK61cnHA30MpcQ3jsW2FlCvmThfRUbTKc
    7JqGsJrTeswCEhCal5EmW1SOB3KDBq6m8MMHbjzx+W7/M5Cn0s5U9scoMn/ITi5u
    hDNC+Z1yYcPUwj89VvyuVQKBgCq9ztxC3vyp7Gf9xJsJ9oG3XfzeKrm2HcBUf2vC
    8txhZWmWpQIeDhRH+i8OvWCwOodCFrxGZ6dWqqX4f/9X4BvFXy/Dv80Ldb6X9O98
    ocYKZ9Rl+wiUbQGuLQd8eTlsoBOMfkDrM6U6pkYzMiLDo2b3nN9DTKVIDbv5Q+tr
    YnbBAoGBAMSqyBvyAsNdek5qvQ5yNqOl6X0HZJ/hp9HYz8gLyqmLWkQTjHUBI9zA
    KwJD5aLsPQCvKrE1oW66XkqI0p1KqZtkfL9ZfFH9A/AhpmEqtTStT62q/Ea21MDU
    AqJIPMXO96oTVY6eorI9BU0cG6n0UvzWZAxNT2UDK07UTB4v7Z8y
    -----END RSA PRIVATE KEY-----
executable_signing_public_key = -----BEGIN PUBLIC KEY-----
    MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAzbuwYnKTTleU2F4zu1gI
    /BolzR74470j6wK7QrQp5b5QkmfevCdX540Ax3mFt6isChrhIXy4LmTvM5SeiCvs
    6su8ro7ZYqtbZjpoBWorESy5VAnnuiuwiAvI7IOPYbSu8e+mHQa893hLYv9q0kQe
    1CtlvQsKir8aQ0YC63UWhrcMcMYCjUaRkWo8JUWoCW/GuZjbuFF0ZZsA13cVSOhF
    g1dxvX4/+FXlnyZQbu2Aex4NU9NGYetl+SP10Uq/c92wZ34ENU3rH8Kw8Tar/dku
    g9+hgrcgt7JRECXmho+WFrKM1H6z1CNe6uEALMfogNl9Mm2jCgavl1wbV92pIKT1
    pQIDAQAB
    -----END PUBLIC KEY-----
""")
    errors = conf.Validate(["Client"])
    self.assertEqual(errors, {})
    self.assertIsInstance(
        conf["Client.executable_signing_public_key"], rdf_crypto.RSAPublicKey
    )

  def testGet(self):
    conf = config_lib.GrrConfigManager()
    conf.DEFINE_string("Section1.foobar", "test", "A test string.")
    conf.DEFINE_string("Section1.foobaz", None, "An empty default string.")
    conf.DEFINE_string("Section1.foobin", "", "An empty default string.")
    conf.initialized = True
    self.assertEqual(conf.Get("Section1.foobar"), "test")
    self.assertEqual(conf.Get("Section1.foobar", default=None), None)
    conf.Initialize(data="""
[Section1]
foobar = X
""")
    self.assertEqual(conf.Get("Section1.foobar", default=None), "X")

    # This not being None is a little surprising, but probably not a big deal
    self.assertEqual(conf.Get("Section1.foobaz"), "")
    self.assertEqual(conf.Get("Section1.foobin"), "")

  def testAddOption(self):
    """Test that we can add options."""
    conf = config_lib.GrrConfigManager()

    conf.DEFINE_string("Section1.foobar", "test", "A test string.")
    conf.DEFINE_string("Section1.test", "test", "A test string.")

    conf.DEFINE_string("Section1.interpolated", "", "An interpolated string.")

    # This entry is not correct - the default is invalid.
    conf.DEFINE_integer("Section1.broken_int", "string", "A test integer.")

    conf.DEFINE_string("Section1.system", None, "The basic operating system.")
    conf.DEFINE_integer("Section1.test_int", 54, "A test integer.")
    conf.DEFINE_list("Section1.test_list", ["a", "b"], "A test integer.")
    conf.DEFINE_list("Section1.test_list2", ["a", "b"], "A test integer.")

    conf.DEFINE_integer("Section2.test_int", None, "A test integer.")
    conf.DEFINE_string("Section2.interpolated", "", "An interpolated string.")

    conf.DEFINE_integer("Section3.test_int", None, "A test integer.")
    conf.DEFINE_string("Section3.interpolated", "", "An interpolated string.")
    conf.Initialize(data="""
[Section1]
foobar = X
test_list = x,y

[Section2]
test_int = 34
interpolated = %(Section1.foobar)Y

[Section3]
test_int = 1
interpolated = %(%(Section1.foobar)|lower)Y

""")

    # The default value is invalid.
    errors = conf.Validate("Section1")
    self.assertIn(
        "Invalid value string for Integer", str(errors["Section1.broken_int"])
    )

    # Section not specified:
    self.assertRaises(config_lib.UnknownOption, conf.__getitem__, "a")

    # Test direct access.
    self.assertEqual(conf["Section1.foobar"], "X")
    self.assertEqual(conf["Section1.test_list"], ["x", "y"])
    self.assertEqual(conf["Section1.test_list2"], ["a", "b"])

    # Test default access.
    self.assertEqual(conf["Section1.test"], "test")

    # Test interpolation with full section name.
    self.assertEqual(conf["Section2.interpolated"], "XY")

    # Check that default values are typed.
    self.assertEqual(conf["Section1.test_int"], 54)

    # Test filter functions.
    self.assertEqual(conf["Section3.interpolated"], "xY")

  def testConstants(self):
    """Test that we can not modify constant values during runtime."""
    conf = config_lib.GrrConfigManager()

    conf.DEFINE_constant_string("Section1.const", "test", "A test string.")

    # We should be able to read this while the config is not initialized.
    self.assertEqual(conf["Section1.const"], "test")

    data = """
[Section1]
const = New string
"""

    # Modifying a constant value in the config file is OK.
    conf.Initialize(data=data)

    # Once the config file is loaded and initialized, modification of constant
    # values is an error.
    self.assertRaises(
        config_lib.ConstModificationError,
        conf.Set,
        "Section1.const",
        "New string",
    )
    self.assertRaises(
        config_lib.ConstModificationError,
        conf.SetRaw,
        "Section1.const",
        "New string",
    )

  @flagsaver.flagsaver(disallow_missing_config_definitions=True)
  def testBadConfigRaises(self):
    conf = config_lib.GrrConfigManager()
    conf.initialized = False
    data = """
Section1.test: 2
"""
    # This config option isn't defined, so it should raise
    with self.assertRaises(config_lib.MissingConfigDefinitionError):
      conf.Initialize(parser=config_parser.YamlConfigFileParser, data=data)

  def testBadFilterRaises(self):
    """Checks that bad filter directive raise."""
    conf = config_lib.GrrConfigManager()
    conf.DEFINE_string(
        "Section1.foo6", "%(somefile@somepackage|resource)", "test"
    )
    conf.DEFINE_string("Section1.foo1", "%(Section1.foo6)/bar", "test")
    conf.Initialize(data="")

    with self.assertRaises(config_lib.InterpolationError) as context:
      _ = conf["Section1.foo1"]

    # Make sure the stringified exception explains the full interpolation chain.
    self.assertIn("%(Section1.foo6)/bar", str(context.exception))

  @flagsaver.flagsaver(disallow_missing_config_definitions=True)
  def testConfigOptionsDefined(self):
    """Test that all config options in use are defined."""
    # We need to use the actual config.CONFIG variable since that is where
    # all the variables are already defined.
    conf = config.CONFIG.MakeNewConfig()

    # Check our actual config validates
    configpath = package.ResourcePath(
        "grr-response-core", "install_data/etc/grr-server.yaml"
    )
    conf.Initialize(filename=configpath)

  def _DefineStringName(self, conf, name):
    conf.DEFINE_string(name, "", "A test.")

  def testUnbalancedParenthesis(self):
    conf = config_lib.GrrConfigManager()
    name_list = [
        "Section1.foobar",
        "Section1.foo",
        "Section1.foo1",
        "Section1.foo2",
        "Section1.foo3",
        "Section1.foo4",
        "Section1.foo5",
        "Section1.foo6",
        "Section1.interpolation1",
        "Section1.interpolation2",
        "Section1.literal",
    ]
    for name in name_list:
      self._DefineStringName(conf, name)

    conf.Initialize(data=r"""
[Section1]
foobar = X
foo = %(Section1.foobar)
foo1 = %(foo

# Unbalanced parenthesis
foo2 = foo)

# Unbalanced parenthesis is ok if escaped.
foo3 = foo\)

# Or if enclosed in a literal block.
foo6 = %{foo)}

foo4 = %{%(hello)}
foo5 = %{hello

# Literal blocks can also appear inside filter interpolations to prevent
# automatic expansions.

# This pull the environment variable "sectionX"
interpolation1 = %(section%(Section1.foobar)|env)

# But this means literally section%(Section1.foo):
interpolation2 = %(section%{%(Section1.foo)}|env)

literal = %{aff4:/C\.(?P<path>.\{1,16\}?)($|/.*)}

""")

    # Test direct access.
    self.assertEqual(conf["Section1.foo"], "X")
    self.assertRaises(
        config_lib.ConfigFormatError, conf.__getitem__, "Section1.foo1"
    )

    self.assertRaises(
        config_lib.ConfigFormatError, conf.__getitem__, "Section1.foo2"
    )

    self.assertEqual(conf["Section1.foo3"], "foo)")

    # Test literal expansion.
    self.assertEqual(conf["Section1.foo4"], "%(hello)")

    self.assertRaises(
        config_lib.ConfigFormatError, conf.__getitem__, "Section1.foo5"
    )

    self.assertEqual(conf["Section1.foo6"], "foo)")

    # The Env filter forces uppercase on args.
    os.environ["sectionX".upper()] = "1"
    os.environ["section%(Section1.foo)".upper()] = "2"

    self.assertEqual(conf["Section1.interpolation1"], "1")
    self.assertEqual(conf["Section1.interpolation2"], "2")

    # Test that Set() escapes - i.e. reading the value back will return exactly
    # the same as we wrote:
    conf.Set("Section1.foo6", "%(Section1.foo3)")
    self.assertEqual(conf["Section1.foo6"], "%(Section1.foo3)")
    self.assertEqual(conf.GetRaw("Section1.foo6"), r"\%(Section1.foo3\)")

    # OTOH when we write it raw, reading it back will interpolate:
    conf.SetRaw("Section1.foo6", "%(Section1.foo3)")
    self.assertEqual(conf["Section1.foo6"], "foo)")

    # A complex regex which gets literally expanded.
    self.assertEqual(
        conf["Section1.literal"], r"aff4:/C\.(?P<path>.{1,16}?)($|/.*)"
    )

  def testDataTypes(self):
    conf = config_lib.GrrConfigManager()
    conf.DEFINE_float("Section1.float", 0, "A float")
    conf.Initialize(
        parser=config_parser.YamlConfigFileParser, data="Section1.float: abc"
    )
    errors = conf.Validate("Section1")
    self.assertIn("Invalid value abc for Float", str(errors["Section1.float"]))

    self.assertRaises(config_lib.ConfigFormatError, conf.Get, "Section1.float")
    conf.Initialize(
        parser=config_parser.YamlConfigFileParser, data="Section1.float: 2"
    )

    # Should have no errors now. Validate should normalize the value to a float.
    self.assertEqual(conf.Validate("Section1"), {})

    self.assertEqual(type(conf.Get("Section1.float")), float)

    conf = config_lib.GrrConfigManager()
    conf.DEFINE_integer("Section1.int", 0, "An integer")
    conf.DEFINE_list("Section1.list", default=[], help="A list")
    conf.DEFINE_list("Section1.list2", default=["a", "2"], help="A list")
    conf.Initialize(
        parser=config_parser.YamlConfigFileParser, data="Section1.int: 2.0"
    )

    errors = conf.Validate("Section1")

    # Floats can not be coerced to an int because that will lose data.
    self.assertIn("Invalid value 2.0 for Integer", str(errors["Section1.int"]))

    # A string can be coerced to an int if it makes sense:
    conf.Initialize(
        parser=config_parser.YamlConfigFileParser, data="Section1.int: '2'"
    )

    conf.Validate("Section1")
    self.assertEqual(type(conf.Get("Section1.int")), int)

    self.assertEqual(type(conf.Get("Section1.list")), list)
    self.assertEqual(conf.Get("Section1.list"), [])

    self.assertEqual(type(conf.Get("Section1.list2")), list)
    self.assertEqual(conf.Get("Section1.list2"), ["a", "2"])

  def _GetNewConf(self):
    conf = config_lib.GrrConfigManager()
    conf.DEFINE_bool("SecondaryFileIncluded", False, "A string")
    conf.DEFINE_bool("TertiaryFileIncluded", False, "A string")
    conf.DEFINE_integer("Section1.int", 0, "An integer")
    conf.DEFINE_context("Client Context")
    return conf

  def _CheckConf(self, conf):
    self.assertTrue(conf.Get("SecondaryFileIncluded"))
    self.assertTrue(conf.Get("TertiaryFileIncluded"))
    self.assertEqual(conf.Get("Section1.int"), 3)

  def testConfigFileInclusion(self):
    one = r"""
Config.includes:
  - 2.yaml

Section1.int: 1
"""
    two = r"""
SecondaryFileIncluded: true
Section1.int: 2
Config.includes:
  - subdir/3.yaml
"""
    three = r"""
TertiaryFileIncluded: true
Section1.int: 3
"""

    with utils.TempDirectory() as temp_dir:
      configone = os.path.join(temp_dir, "1.yaml")
      configtwo = os.path.join(temp_dir, "2.yaml")
      subdir = os.path.join(temp_dir, "subdir")
      os.makedirs(subdir)
      configthree = os.path.join(subdir, "3.yaml")
      with io.open(configone, "w") as fd:
        fd.write(one)

      with io.open(configtwo, "w") as fd:
        fd.write(two)

      with io.open(configthree, "w") as fd:
        fd.write(three)

      # Using filename
      conf = self._GetNewConf()
      conf.Initialize(
          parser=config_parser.YamlConfigFileParser, filename=configone
      )
      self._CheckConf(conf)

      # Using fd with no fd.name should raise because there is no way to resolve
      # the relative path.
      conf = self._GetNewConf()
      fd = io.BytesIO(one.encode("utf-8"))
      self.assertRaises(
          config_lib.ConfigFileNotFound,
          conf.Initialize,
          parser=config_parser.YamlConfigFileParser,
          fd=fd,
      )

      # Using data
      conf = self._GetNewConf()
      self.assertRaises(
          config_lib.ConfigFileNotFound,
          conf.Initialize,
          parser=config_parser.YamlConfigFileParser,
          data=one,
      )

  def testConfigFileInclusionCanBeTurnedOff(self):
    one = r"""
Config.includes:
  - 2.yaml

Section1.int: 1
"""
    two = r"""
SecondaryFileIncluded: true
Section1.int: 2
"""

    with utils.TempDirectory() as temp_dir:
      configone = os.path.join(temp_dir, "1.yaml")
      configtwo = os.path.join(temp_dir, "2.yaml")
      with io.open(configone, "w") as fd:
        fd.write(one)
      with io.open(configtwo, "w") as fd:
        fd.write(two)

      # Using filename
      conf = self._GetNewConf()
      conf.Initialize(
          parser=config_parser.YamlConfigFileParser,
          filename=configone,
          process_includes=False,
      )

      self.assertFalse(conf.Get("SecondaryFileIncluded"))
      self.assertEqual(conf.Get("Section1.int"), 1)

  def testConfigFileIncludeAbsolutePaths(self):
    one = r"""
Section1.int: 1
"""
    with utils.TempDirectory() as temp_dir:
      configone = os.path.join(temp_dir, "1.yaml")
      with io.open(configone, "w") as fd:
        fd.write(one)

      absolute_include = r"""
Config.includes:
  - %s

Section1.int: 2
""" % configone

      conf = self._GetNewConf()
      conf.Initialize(
          parser=config_parser.YamlConfigFileParser, data=absolute_include
      )
      self.assertEqual(conf["Section1.int"], 1)

      relative_include = r"""
Config.includes:
  - 1.yaml

Section1.int: 2
"""
      conf = self._GetNewConf()
      # Can not include a relative path from config without a filename.
      self.assertRaises(
          config_lib.ConfigFileNotFound,
          conf.Initialize,
          parser=config_parser.YamlConfigFileParser,
          data=relative_include,
      )

      # If we write it to a file it should work though.
      configtwo = os.path.join(temp_dir, "2.yaml")
      with io.open(configtwo, "w") as fd:
        fd.write(relative_include)

      conf.Initialize(
          parser=config_parser.YamlConfigFileParser, filename=configtwo
      )
      self.assertEqual(conf["Section1.int"], 1)

  def testConfigFileInclusionWindowsPaths(self):
    one = r"""
Config.includes:
  - 2.yaml

Section1.int: 1
"""
    two = r"""
Section1.int: 2
SecondaryFileIncluded: true
"""
    config_path = "C:\\Windows\\System32\\GRR"

    def MockedWindowsOpen(filename, _=None):
      basename = ntpath.basename(filename)
      dirname = ntpath.dirname(filename)

      # Make sure we only try to open files from this directory.
      if dirname != config_path:
        raise IOError("Tried to open wrong file %s" % filename)

      if basename == "1.yaml":
        return io.BytesIO(one.encode("utf-8"))

      if basename == "2.yaml":
        return io.BytesIO(two.encode("utf-8"))

      raise IOError("File not found %s" % filename)

    # TODO(user): this kind of mocking is a questionable practice at best.
    # We have Windows-specific tests and should use them for this kind of
    # testing.
    #
    # We need to also use the nt path manipulation modules.
    with utils.MultiStubber(
        (io, "open", MockedWindowsOpen), (os, "path", ntpath)
    ):
      conf = self._GetNewConf()
      conf.Initialize(filename=ntpath.join(config_path, "1.yaml"))
      self.assertEqual(conf["Section1.int"], 2)
      self.assertEqual(conf["SecondaryFileIncluded"], True)

  def testConfigFileInclusionWithContext(self):
    one = r"""
Client Context:
  Config.includes:
    - 2.yaml

Section1.int: 1
"""
    two = r"""
Section1.int: 2
SecondaryFileIncluded: true
"""
    with utils.TempDirectory() as temp_dir:
      configone = os.path.join(temp_dir, "1.yaml")
      configtwo = os.path.join(temp_dir, "2.yaml")
      with io.open(configone, "w") as fd:
        fd.write(one)

      with io.open(configtwo, "w") as fd:
        fd.write(two)

      # Without specifying the context the includes are not processed.
      conf = self._GetNewConf()
      conf.Initialize(
          parser=config_parser.YamlConfigFileParser, filename=configone
      )
      self.assertEqual(conf["Section1.int"], 1)

      # Only one config is loaded.
      self.assertEqual(conf.files, [configone])

      # Now we specify the context.
      conf = self._GetNewConf()
      conf.AddContext("Client Context")
      conf.Initialize(
          parser=config_parser.YamlConfigFileParser, filename=configone
      )

      # Both config files were loaded. Note that load order is important and
      # well defined.
      self.assertEqual(conf.files, [configone, configtwo])
      self.assertEqual(conf["Section1.int"], 2)

  def testMatchBuildContext(self):
    context = """
Test1 Context:
  ClientBuilder.target_platforms:
    - linux_amd64_deb
    - linux_i386_deb
    - windows_amd64_exe

Test2 Context:

Test3 Context:
  ClientBuilder.target_platforms:
    - linux_amd64_deb
    - windows_i386_exe
"""
    conf = config.CONFIG.MakeNewConfig()
    conf.DEFINE_context("Test1 Context")
    conf.DEFINE_context("Test2 Context")
    conf.DEFINE_context("Test3 Context")
    conf.Initialize(parser=config_parser.YamlConfigFileParser, data=context)
    conf.AddContext("Test1 Context")
    result_map = [
        (("linux", "amd64", "deb"), True),
        (("linux", "i386", "deb"), True),
        (("windows", "amd64", "exe"), True),
        (("windows", "i386", "exe"), False),
    ]
    for result in result_map:
      self.assertEqual(conf.MatchBuildContext(*result[0]), result[1])

  def testMatchBuildContextError(self):
    """Raise because the same target was listed twice."""
    context = """
Test1 Context:
  ClientBuilder.target_platforms:
    - linux_amd64_deb
    - linux_i386_deb
    - linux_amd64_deb
    - windows_amd64_exe
"""
    conf = config.CONFIG.MakeNewConfig()
    conf.DEFINE_context("Test1 Context")
    conf.Initialize(parser=config_parser.YamlConfigFileParser, data=context)
    conf.AddContext("Test1 Context")
    with self.assertRaises(type_info.TypeValueError):
      conf.MatchBuildContext("linux", "amd64", "deb")

  def testNoUnicodeWriting(self):
    conf = config.CONFIG.MakeNewConfig()
    config_file = os.path.join(self.temp_dir, "writeback.yaml")
    conf.SetWriteBack(config_file)
    conf.DEFINE_string("NewSection1.new_option1", "Default Value", "Help")
    conf.Set(str("NewSection1.new_option1"), "New Value1")
    conf.Write()

    data = io.open(config_file).read()
    self.assertNotIn("!!python/unicode", data)

  def testRenameOnWritebackFailure(self):
    conf = config.CONFIG.MakeNewConfig()
    writeback_file = os.path.join(self.temp_dir, "writeback.yaml")
    with io.open(writeback_file, "w") as f:
      f.write("This is a bad line of yaml{[(\n")
      f.close()

    self.assertRaises(AttributeError, conf.SetWriteBack, writeback_file)
    self.assertTrue(os.path.isfile(writeback_file + ".bak"))

  def testNoRenameOfReadProtectedFile(self):
    """Don't rename config files we don't have permission to read."""
    writeback_file = os.path.join(self.temp_dir, "writeback.yaml")
    with io.open(writeback_file, mode="w", encoding="utf-8") as f:
      f.write("")

    # Remove all permissions except user write.
    os.chmod(writeback_file, stat.S_IWUSR)

    conf = config.CONFIG.MakeNewConfig()
    conf.SetWriteBack(writeback_file)

    # File is still in the same place
    self.assertTrue(os.path.isfile(writeback_file))


def main(argv):
  test_lib.main(argv)


if __name__ == "__main__":
  app.run(main)



#!/usr/bin/env python
"""A module with utilities for optimized pseudo-random number generation."""

from collections.abc import Callable
import os
import struct
import threading

_random_buffer_size: int = 1024
_random_buffer: list[int] = []
_mutex = threading.Lock()


def UInt16() -> int:
  """Returns a pseudo-random 16-bit unsigned integer."""
  return UInt32() & 0xFFFF


def PositiveUInt16() -> int:
  """Returns a pseudo-random 16-bit non-zero unsigned integer."""
  return _Positive(UInt16)


def UInt32() -> int:
  """Returns a pseudo-random 32-bit unsigned integer."""
  with _mutex:
    try:
      return _random_buffer.pop()
    except IndexError:
      data = os.urandom(struct.calcsize("=L") * _random_buffer_size)
      _random_buffer.extend(
          struct.unpack("=" + "L" * _random_buffer_size, data)
      )
      return _random_buffer.pop()


def PositiveUInt32() -> int:
  """Returns a pseudo-random 32-bit non-zero unsigned integer."""
  return _Positive(UInt32)


def UInt64() -> int:
  """Returns a pseudo-random 64-bit unsigned integer."""
  return (UInt32() << 32) | UInt32()


def Id64() -> int:
  """Returns a pseudo-random, positive, non-zero 64-bit int usable as ID.

  The returned int is guaranteed to have at least one of the most significant 32
  bits set to 1. Effectively, all IDs generated by this function are
  >= 0x100000000. This is due to backwards compatibility: Legacy, 32-bit IDs can
  be identified by having 0 in the 32 most significant bits. Legacy IDs are
  serialized as 8-character hexadecimal, whereas new IDs are serialized as 16-
  character hexadecimal string.
  """
  return (PositiveUInt32() << 32) | UInt32()


def _Positive(rng: Callable[[], int]) -> int:
  while True:
    result = rng()
    if result > 0:
      return result



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import client_fs as rdf_client_fs
from grr_response_proto import flows_pb2
from grr_response_proto import jobs_pb2
from grr_response_proto import sysinfo_pb2


def ToProtoFilesystem(rdf: rdf_client_fs.Filesystem) -> sysinfo_pb2.Filesystem:
  return rdf.AsPrimitiveProto()


def ToRDFFilesystem(proto: sysinfo_pb2.Filesystem) -> rdf_client_fs.Filesystem:
  return rdf_client_fs.Filesystem.FromSerializedBytes(proto.SerializeToString())


def ToProtoWindowsVolume(
    rdf: rdf_client_fs.WindowsVolume,
) -> sysinfo_pb2.WindowsVolume:
  return rdf.AsPrimitiveProto()


def ToRDFWindowsVolume(
    proto: sysinfo_pb2.WindowsVolume,
) -> rdf_client_fs.WindowsVolume:
  return rdf_client_fs.WindowsVolume.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoUnixVolume(rdf: rdf_client_fs.UnixVolume) -> sysinfo_pb2.UnixVolume:
  return rdf.AsPrimitiveProto()


def ToRDFUnixVolume(proto: sysinfo_pb2.UnixVolume) -> rdf_client_fs.UnixVolume:
  return rdf_client_fs.UnixVolume.FromSerializedBytes(proto.SerializeToString())


def ToProtoVolume(rdf: rdf_client_fs.Volume) -> sysinfo_pb2.Volume:
  return rdf.AsPrimitiveProto()


def ToRDFVolume(proto: sysinfo_pb2.Volume) -> rdf_client_fs.Volume:
  return rdf_client_fs.Volume.FromSerializedBytes(proto.SerializeToString())


def ToProtoDiskUsage(rdf: rdf_client_fs.DiskUsage) -> sysinfo_pb2.DiskUsage:
  return rdf.AsPrimitiveProto()


def ToRDFDiskUsage(proto: sysinfo_pb2.DiskUsage) -> rdf_client_fs.DiskUsage:
  return rdf_client_fs.DiskUsage.FromSerializedBytes(proto.SerializeToString())


def ToProtoExtAttr(rdf: rdf_client_fs.ExtAttr) -> jobs_pb2.StatEntry.ExtAttr:
  return rdf.AsPrimitiveProto()


def ToRDFExtAttr(proto: jobs_pb2.StatEntry.ExtAttr) -> rdf_client_fs.ExtAttr:
  return rdf_client_fs.ExtAttr.FromSerializedBytes(proto.SerializeToString())


def ToProtoStatEntry(rdf: rdf_client_fs.StatEntry) -> jobs_pb2.StatEntry:
  return rdf.AsPrimitiveProto()


def ToRDFStatEntry(proto: jobs_pb2.StatEntry) -> rdf_client_fs.StatEntry:
  return rdf_client_fs.StatEntry.FromSerializedBytes(proto.SerializeToString())


def ToProtoFindSpec(rdf: rdf_client_fs.FindSpec) -> jobs_pb2.FindSpec:
  return rdf.AsPrimitiveProto()


def ToRDFFindSpec(proto: jobs_pb2.FindSpec) -> rdf_client_fs.FindSpec:
  return rdf_client_fs.FindSpec.FromSerializedBytes(proto.SerializeToString())


def ToProtoBareGrepSpec(
    rdf: rdf_client_fs.BareGrepSpec,
) -> flows_pb2.BareGrepSpec:
  return rdf.AsPrimitiveProto()


def ToRDFBareGrepSpec(
    proto: flows_pb2.BareGrepSpec,
) -> rdf_client_fs.BareGrepSpec:
  return rdf_client_fs.BareGrepSpec.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoGrepSpec(rdf: rdf_client_fs.GrepSpec) -> jobs_pb2.GrepSpec:
  return rdf.AsPrimitiveProto()


def ToRDFGrepSpec(proto: jobs_pb2.GrepSpec) -> rdf_client_fs.GrepSpec:
  return rdf_client_fs.GrepSpec.FromSerializedBytes(proto.SerializeToString())


def ToProtoBlobImageChunkDescriptor(
    rdf: rdf_client_fs.BlobImageChunkDescriptor,
) -> jobs_pb2.BlobImageChunkDescriptor:
  return rdf.AsPrimitiveProto()


def ToRDFBlobImageChunkDescriptor(
    proto: jobs_pb2.BlobImageChunkDescriptor,
) -> rdf_client_fs.BlobImageChunkDescriptor:
  return rdf_client_fs.BlobImageChunkDescriptor.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoBlobImageDescriptor(
    rdf: rdf_client_fs.BlobImageDescriptor,
) -> jobs_pb2.BlobImageDescriptor:
  return rdf.AsPrimitiveProto()


def ToRDFBlobImageDescriptor(
    proto: jobs_pb2.BlobImageDescriptor,
) -> rdf_client_fs.BlobImageDescriptor:
  return rdf_client_fs.BlobImageDescriptor.FromSerializedBytes(
      proto.SerializeToString()
  )
